# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# -----------------------------------------------------------------------------
import re
import toml
import sys
#
def system_exit(msg) :
   # assert False, msg
   sys.exit(msg)
# -----------------------------------------------------------------------------
default_dict = dict()
#
'''
{xrst_begin config_file user}
{xrst_spell
   booleans
   conf
   epilog
   frist
   prolog
   pyenchant
   pyspellchecker
   rtd
   toc
   toml
}

.. _toml file: https://toml.io/en/

Configuration File for xrst
###########################
A `toml file`_ is used to configure xrst.

#. The location of this file is specified by the xrst
   :ref:`run_xrst@config_file` argument.
#. This file is a sequence of toml tables,
   if a table can only have one entry, the entry is named data.
#. Each table has a default value that is used when the table
   is not present in the toml file.
   All of the defaults below will be the same in future versions
   of xrst with the possible exception of the html theme options
   :ref:`config_file@html_theme_options@Default` .
#. All of the entries in a table have the same type
   as its corresponding default.
   If an entry has components, all of the comments have the same
   type as the default components.
   The html theme options default is one exception to this rule; see
   :ref:`config_file@html_theme_options` .
{xrst_comment --------------------------------------------------------------- }

project_name
************
The only value in this table is the name of this project.

Default
=======
{xrst_code toml}
[project_name]
data = 'project'
{xrst_code}
{xrst_suspend}'''
default_dict['project_name'] = { 'data' : 'project' }
'''{xrst_resume}

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_PROJECT_NAME
   # END_PROJECT_NAME
}
{xrst_comment --------------------------------------------------------------- }

directory
*********
This table specifies the locations of some xrst directories:

project_directory
=================
This *project_directory* can be an absolute path,
or a path relative to the location where the
:ref:`run_xrst@config_file` is located.
All of the other directories are specified relative to this directory.
The other directories may have ``../`` in their specifications; i.e.,
they do not need to be sub-directories of the project directory.
This directory must exists when xrst is run.

html_directory
==============
This is the directory, relative to the *project_directory*,
where the output files are stored when
:ref:`run_xrst@target` is ``html`` .
If *target* is html and this directory does not exist, it will be created.

tex_directory
=============
This is the directory, relative to the *project_directory*,
where the output files are stored when
:ref:`run_xrst@target` is ``tex`` .
If *target* is tex and this directory does not exist, it will be created.

rst_directory
=============
This is the directory, relative to the *project_directory*,
where xrst writes the rst files it extracts from the source code.
For each :ref:`begin_cmd@page_name` , the file
If this  directory does not exist, it will be created.

|space| *rst_directory*\ /\ *page_name*\ ``.rst``

is the RST file for the corresponding page. There is one exception
to this rule. If *page_name* ends with ``.rst``, the extra ``.rst``
is not added at the end.

Other Generated Files
---------------------
See :ref:`auto_file-title` for the other files generated by ``xrst``
and placed in the *rst_directory*.

Default
=======
Note that '.' denotes the directory where
:ref:`xrst <run_xrst-name>` is run.

{xrst_code toml}
[directory]
project_directory  = '.'
rst_directory      = 'build/rst'
html_directory     = 'build/html'
tex_directory      = 'build/tex'
{xrst_code}
{xrst_suspend}'''
default_dict['directory'] = {
   'project_directory' : '.'    ,
   'rst_directory'     : 'build/rst'  ,
   'html_directory'    : 'build/html' ,
   'tex_directory'     : 'build/tex'  ,
}
'''{xrst_resume}

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_DIRECTORY
   # END_DIRECTORY
}
{xrst_comment --------------------------------------------------------------- }

root_file
*********
This table maps each :ref:`begin_cmd@group_name`
to its root (top) xrst input file.
These file names are relative to the
:ref:`config_file@directory@project_directory` .
Multiple groups can use the same root file.

Default
=======
{xrst_code toml}
[root_file]
default = 'project.xrst'
{xrst_code}
{xrst_suspend}'''
default_dict['root_file'] = { 'default' : 'project.xrst' }
'''{xrst_resume}

Note that ``default`` corresponds to the
:ref:`begin_cmd@group_name@Default Group` and ``project.xrst``
is the default root file.

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_ROOT_FILE
   # END_ROOT_FILE
}
{xrst_comment --------------------------------------------------------------- }

spell_package
*************
This str is either 'pyspellchecker' or 'pyenchant' .
These are the only spell checkers support so far.
If you use one spell checker the other one need not be installed
on your system.

Default
=======
{xrst_code toml}
[spell_package]
data = 'pyspellchecker'
{xrst_code}
{xrst_suspend}'''
default_dict['spell_package'] = { 'data' : 'pyspellchecker' }
'''{xrst_resume}

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_SPELL_PACKAGE
   # END_SPELL_PACKAGE
}
{xrst_comment --------------------------------------------------------------- }

input_files
***********
This table is used to list the files that should be include in the
documentation if they have a :ref:`begin command<begin_cmd-name>`
for a group in :ref:`run_xrst@group_list` .
The only value in this table is a list of program commands.

Each program command is a list of strings.
The list is the program to execute followed by
the program's command line arguments in order.
The standard output for the program should be a space separated
list of file names to be checked.
If a file name has spaces in it, it should be surrounded by single
or double quotes.
The single and double quotes are not part of the file name.
If this list of strings is empty, no files are checked.

Each program is execute, in order, with the *project_directory*
as the current working directory.
The frist program to return without an error is used for the list of files.
The intention here is that different programs
may be intended for different systems.

If the list of program commands is empty,
no checking of input file list is done.
If the list of program commands is non-empty,
and none of the program commands succeed,
a warning is printed.

Default
=======
{xrst_code toml}
[input_files]
data = [
   [ 'git', 'ls-files' ]
]
{xrst_code}
{xrst_suspend}'''
default_dict['input_files'] = { 'data' : [
   [ 'git', 'ls-files'] ,
] }
'''{xrst_resume}

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_INPUT_FILES
   # END_INPUT_FILES
}

input_files.sh
==============
{xrst_literal
   bin/input_files.sh
}
{xrst_comment --------------------------------------------------------------- }

html_theme_options
******************
This table maps the :ref:`run_xrst@html_theme` to
the corresponding html_theme_options.
This enable building documentation with different themes without
having to change the toml file between builds.
If *html_theme* does not appear in this mapping, no options are set.

For each theme, the corresponding options are in a dictionary.
The keys of this dictionary are strings, but the values do not
necessarily have the same type as he default case.
In the default below, these values are integers and booleans.

Default
=======
The html_theme_options default value below my change in the future.
{xrst_code toml}
[html_theme_options.sphinx_book_theme]
show_toc_level = 4

[html_theme_options.sphinx_rtd_theme]
navigation_depth = -1
titles_only        = true
{xrst_code}
{xrst_suspend}'''
default_dict['html_theme_options'] = {
   'sphinx_book_theme' : {
      'show_toc_level' : 4
   },
   'sphinx_rtd_theme'  : {
      'navigation_depth' : -1   ,
      'titles_only'      : True ,
   },
}
'''{xrst_resume}

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_HTML_THEME_OPTIONS
   # END_HTML_THEME_OPTIONS
}
{xrst_comment --------------------------------------------------------------- }

include_all
***********
This table is used to create input that is included
at the beginning or end of every page.
This table has the following keys:

rst_epilog
==========
The value corresponding to this key is a set of rst commands
that get included at the end of every section.
This is the same as the rst_epilog variable in the sphinx conf.py file.

rst_prolog
==========
The value corresponding to this key is a set of rst commands
that get included at the beginning of every section.
This is the same as the rst_epilog variable in the sphinx conf.py file
except that the latex macros are added at the end when
:ref:`run_xrst@target` is html.

latex_macro
===========
The value corresponding to this key is a list of latex macros.
If :ref:`run_xrst@target` is ``html``, these macros get included at the
beginning of every page using the sphinx ``:math`` role in the
rst_epilog variable in the sphinx conf.py file.
Otherwise *target* is ``tex`` and these macros get included once
at the beginning of the corresponding latex document.
It either case they can be used by every page in the documentation.

Default
=======
{xrst_code toml}
[include_all]
rst_epilog = ''
rst_prolog = ''
latex_macro  = []
{xrst_code}
{xrst_suspend}'''
default_dict['include_all'] = {
      'rst_epilog'   : ''     ,
      'rst_prolog'   : ''     ,
      'latex_macro'  : list() ,
}
'''{xrst_resume}


Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_INCLUDE_ALL
   # END_INCLUDE_ALL
}
{xrst_comment --------------------------------------------------------------- }

project_dictionary
******************
The only value in this table is a list of strings.
Each string contains a newline separated list of words.
Leading and trailing white space is not part of each word.
These special words are not considered spelling errors for the entire project.
Special words, for a particular page, are specified using the
:ref:`spell command<spell_cmd-name>`.

Default
=======
{xrst_code toml}
[project_dictionary]
data = []
{xrst_code}
{xrst_suspend}'''
default_dict['project_dictionary'] = { 'data' : list() }
'''{xrst_resume}

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_PROJECT_DICTIONARY
   # END_PROJECT_DICTIONARY
}
{xrst_comment --------------------------------------------------------------- }

not_in_index
************
The only value in this table is a list of strings.
Each string  contains a newline separated list of patterns.
Leading and trailing white space is not part of each pattern.
These are python regular expression patterns for heading tokens
that are not included in the index.
A heading token is any sequence of non white space characters
with upper case letters converted to lower case.
For example, a heading might contain the token ``The`` but you
might not want to include ``the`` as a entry in the :ref:`genindex`.
In this case you could have a line containing just ``the`` .
For another example, you might want to exclude all tokens that are
integer numbers.
In this case you could have a line containing just ``[0-9]*`` .

Default
=======
{xrst_code toml}
[not_in_index]
data = []
{xrst_code}
{xrst_suspend}'''
default_dict['not_in_index'] = { 'data' : list() }
'''{xrst_resume}

Example
=======
{xrst_literal
   xrst.toml
   # BEGIN_NOT_IN_INDEX
   # END_NOT_IN_INDEX
}

{xrst_end config_file}
-----------------------------------------------------------------------------
'''
def iterable2string(iterable) :
   iterable = sorted( iterable )
   string = ''
   for item in iterable :
      if string != '' :
         string += ', '
      string += str(item)
   return string
#
# {xrst_begin get_conf_dict dev}
# {xrst_spell
#     conf
#     config
#     toml
# }
# {xrst_comment_ch #}
#
# Get Configuration Dictionary
# ############################
# This routine is called before the current working directory is changed to
# the *project_directory* (because it determines the project directory)
# so it cannot use the xrst :ref:`system_exit-name` .
#
# config_file
# ***********
# is the location of the :ref:`run_xrst@config_file` specified on
# the xrst command line.
#
# conf_dict
# *********
# is the python dictionary corresponding to the toml file with the defaults
# filled in. All of the values in the dictionary have been check for
# the proper type. This includes recursive checking; e.g. a list is checked
# to make sure its elements have the proper type.
#
# {xrst_code py}
# conf_dict =
def get_conf_dict(config_file) :
   assert type(config_file) == str
   # {xrst_code}
   # {xrst_literal
   #     BEGIN_RETURN
   #     END_RETURN
   # }
   # {xrst_end get_conf_dict}
   #
   # msg
   msg  = f'config_file = {config_file}\n'
   #
   # conf_dict
   file_obj  = open(config_file, 'r')
   file_data = file_obj.read()
   conf_dict = toml.loads(file_data)
   #
   # check top level keys in config_file
   for table in conf_dict :
      if table not in default_dict :
         msg += 'This file has the unexpected table:\n'
         msg += f'   {table}\n'
         msg += 'The value table names are:\n'
         msg += iterable2string( default_dict.keys() )
         system_exit(msg)
   #
   # conf_dict
   # also check that all the top keys the correct type
   for table in default_dict :
      default = default_dict[table]
      if table not in conf_dict :
         conf_dict[table] = default
      else :
         table_dict = conf_dict[table]
         if type(table_dict) != dict :
            msg += f'Table {table} is not a python dictionary\n'
            msg += 'it has python type '+ str( type(table_dict) )
            system_exit(msg)
   #
   # directory
   table_dict = conf_dict['directory']
   valid_set = {
      'project_directory',
      'rst_directory',
      'html_directory',
      'tex_directory',
   }
   if set( table_dict.keys() ) != valid_set :
      msg += 'The directory has the following keys:\n'
      msg += iterable2string( table_dict.keys() ) + '\n'
      msg += 'Expected the following keys:\n'
      msg += iterable2string( valid_set ) + '\n'
      system_exit(msg)
   for key in table_dict :
      value = table_dict[key]
      if type(value) != str :
         msg += f'directory.{key} has python type ' + str(type(value))
         system_exit(msg)
   #
   # spell_package
   value = conf_dict['spell_package']['data']
   if value not in [ 'pyspellchecker', 'pyenchant' ] :
      msg += 'spell_package has is not pyspellchecker or pyenchant'
      system_exit(msg)
   #
   # root_file
   p_group_name = re.compile( r'[a-z]+' )
   table_dict = conf_dict['root_file']
   for key in table_dict :
      if key == '' :
         msg += 'root_file uses the empty group name\n'
         msg += 'Use default for the empty group name'
         system_exit(msg)
      m_group_name = p_group_name.fullmatch(key)
      if m_group_name == None :
         msg += f'root_file: "{key}" is not a valid group name\n'
         msg += 'A valid group name is a sequence of the letters a-z.'
         system_exit(msg)
      value = table_dict[key]
      if type(value) != str :
         msg += f'root_file.{key} value has python type '
         msg += str(type(value)) + '\n'
         msg += 'Expected it to have type ' + str( str )
         system_exit(msg)
   #
   # input_files
   table_dict = conf_dict['input_files']
   value      = table_dict['data']
   if type(value) != list :
         msg += f'input_files.data has python type '
         msg += str(type(value)) + '\n'
         msg += 'Expected it to have type ' + str( list )
         system_exit(msg)
   for (index1, entry1) in enumerate(value) :
      if type(entry1) != list :
         msg += f'input_files.data[{index1}] has python type '
         msg += str(type(entry1)) + '\n'
         msg += 'Expected it to have type ' + str( list )
         system_exit(msg)
      for (index2, entry2) in enumerate(entry1) :
         if type(entry2) != str :
            msg += f'input_files.data[{index1}][{index2}] has python type '
            msg += str(type(entry1)) + '\n'
            msg += 'Expected it to have type ' + str( str )
            system_exit(msg)
   #
   # html_theme_options
   table_dict = conf_dict['html_theme_options']
   for key in table_dict :
      value = table_dict[key]
      try :
         dict( value )
      except :
         msg += f'html_theme_options.{key} has python type ' + str(type(value))
         msg += '\nExpected it to have type ' + str( dict )
         system_exit(msg)
   #
   # include_all
   table_dict = conf_dict['include_all']
   valid_set  = {
      'rst_epilog',
      'rst_prolog',
      'latex_macro'
   }
   if set(table_dict.keys()) != valid_set :
      msg += 'The include_all has the following keys:\n'
      msg += iterable2string( table_dict.keys() ) + '\n'
      msg += 'Expected the following keys:\n'
      msg += iterable2string( valid_set ) + '\n'
      system_exit(msg)
   for key in [ 'rst_epilog', 'rst_prolog' ] :
      value = table_dict[key]
      if type( value ) != str :
         msg += f'include_all.{key} has python type '
         msg += str(type(value)) + '\n'
         msg += 'Expected it to have type ' + str( str )
         system_exit(msg)
   value = table_dict['latex_macro']
   if type( value ) != list :
      msg += f'include_all.latex_macro has python type '
      msg += str(type(value)) + '\n'
      msg += 'Expected it to have type ' + str( list )
      system_exit(msg)
   for (index, entry) in enumerate(value) :
      if type(entry) != str :
         msg += f'include_all.latex_macro[{index}] has python type '
         msg += str(type(entry)) + '\n'
         msg += 'Expected it to have type ' + str( str )
         system_exit(msg)
   #
   # project_dictionary, not_in_index
   for table in [ 'project_dictionary', 'not_in_index' ] :
      value = conf_dict[table]['data']
      if type(value) != list :
            msg += f'project_dictionary.data has python type '
            msg += str(type(value)) + '\n'
            msg += 'Expected it to have type ' + str( list )
            system_exit(msg)
      for (index, entry) in enumerate(value) :
         if type(entry) != str :
            msg += f'output_direcory.data[{index}] has python type '
            msg += str(type(entry)) + '\n'
            msg += 'Expected it to have type ' + str( str )
            system_exit(msg)
   #
   # BEGIN_RETURN
   assert type(conf_dict) == dict
   return conf_dict
   # END_RETURN



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin run_xrst user}
{xrst_spell
   config
   dev
   furo
   github
   jax
   pdf
   pyspellchecker
   rtd
   thet
   toc
   toml
   txt
   wil
}

Run Extract Sphinx RST And Sphinx
#################################

Syntax
******
It may be helpful to remove the :ref:`config_file@directory@html_directory`
before running the command below.
This will check for error messages that are not repeated due
to caching the results of previous builds.

| ``xrst`` \\
| |tab| [ ``--version`` ] \\
| |tab| [ ``--local_toc`` ] \\
| |tab| [ ``--page_source`` ] \\
| |tab| [ ``--rst_line_numbers`` ] \\
| |tab| [ ``--rst_only`` ] \\
| |tab| [ ``--index_page_name`` *index_page_name* ] \\
| |tab| [ ``--config_file``     *config_file* ] \\
| |tab| [ ``--html_theme``      *html_theme* ] \\
| |tab| [ ``--target``          *target* ]  \\
| |tab| [ ``--group_list``      *group_name_1* *group_name_2* ... ] \\
| |tab| [ ``--rename_group``    *old_group_name* *new_group_name* ] \\
| |tab| [ ``--replace_spell_commands`` ] \\
{xrst_comment --------------------------------------------------------------- }

version
*******
If ``--version`` is present on the command line,
the version of xrst is printed and none of the other arguments matter.
{xrst_comment --------------------------------------------------------------- }

local_toc
*********
If this option is present on the command line,
a table of contents for the Headings in the current page
is included at the top of every page.
The page name and page title are not in this table of contents.

Some :ref:`html themes<run_xrst@html_theme>` include this information
on a side bar; e.g., ``furo`` and ``sphinx_book_theme`` .
{xrst_comment --------------------------------------------------------------- }

page_source
***********
If this option is present and *target* is ``html`` ,
a link to the xrst source code is included at the top of each page.
Some :ref:`html themes<run_xrst@html_theme>` include this link; e.g.,
``sphinx_rtd_theme`` .

If this option is present and *target* is ``tex`` ,
the xrst source code file is reported at the beginning of each page.
{xrst_comment --------------------------------------------------------------- }

rst_line_numbers
****************
Normally sphinx error and warning messages are reported using line numbers
in the xrst source code files.
If this option is present, these messages are reported
using the line numbers in the RST files created by xrst.
In addition the :ref:`run_xrst@page_source` links to the rst files,
instead of the xrst source files.
This may be helpful if you have an error or warning for a sphinx command
and it does not make sense using xrst source code line numbers.
It is also helpful for determining if an incorrect line number is due to
sphinx or xrst.
{xrst_comment --------------------------------------------------------------- }

rst_only
********
Normally, after extraction the RST files,
xrst automatically runs sphinx to produce the target output (html or tex).
If this option is present, sphinx not run.
Only the rst files, and their corresponding sources,
are generated; i.e.,

| |tab| :ref:`config_file@directory@rst_directory`/\*.rst
| |tab| *rst_directory*\ /_sources/\*.txt

This may be useful when creating rst files for uses else where; e.g.,
for use with `Read the Docs <https://docs.readthedocs.io>`_ .
The sphinx commands are printed after xrst finishes and can be executed
by hand.
This may be useful if there is a problem during these commands.
{xrst_comment --------------------------------------------------------------- }

index_page_name
***************
This option has no effect when *target* is ``tex`` .
If *target* is ``html``,
the file ``index.html`` in the
:ref:`config_file@directory@html_directory` will be a redirect
to the page specified by *index_page_name* .
If this option is not present, ``index.html`` wil be a redirect
to the root of the documentation tree.
{xrst_comment --------------------------------------------------------------- }

config_file
***********
The command line argument *config_file* specifies the location of the
:ref:`config_file-name` for this project.
This can be an absolute path or
relative to the directory where :ref:`xrst<run_xrst-name>` is run.

xrst.toml
=========
If *config_file* is not present on the command line,
the default value ``xrst.toml`` is used for *config_file* .
{xrst_comment --------------------------------------------------------------- }

html_theme
**********
This the html_theme_ that is used.
The default value for *html_theme* is ``furo`` .

.. _html_theme: https://sphinx-themes.org/

Theme Choices
=============
The following is a list of some themes that work well with the
default settings in :ref:`config_file@html_theme_options` .
If you have a theme together with html_theme_options
that work well with xrst,
please post an issue on github so that it can be added to the list below.

{xrst_spell_off}
.. csv-table:: Sphinx Themes
   :header: name,  local_toc

   sphinx_rtd_theme,     yes
   furo,                 no
   sphinx_book_theme,    no
   pydata_sphinx_theme,  no
   piccolo_theme,        no
{xrst_spell_on}

sphinx_rtd_theme
================
The sphinx_rtd theme builds faster than some of the other themes,
so it is suggested to use it for testing (with the ``--local_toc`` option).
A special modification is made to this theme when *target* is html,
so that it displays wider than its normal limit.
This modification may be removed in the future.
{xrst_comment --------------------------------------------------------------- }

target
******
The optional command line argument *target* must be ``html`` or ``tex``.
It specifies the type of type output you plan to generate using sphinx.
Note thet :ref:`config_file@directory@html_directory` and
:ref:`config_file@directory@tex_directory` will determine the location
of the corresponding output files.
The default value for *target* is ``html`` .

tex
===
If you choose this target, xrst will create the file
*project_name*\ ``.tex`` in the :ref:`config_file@directory@tex_directory` .
There are two reasons to build this file.
One is to create the file *project_name*\ ``.pdf``
which is a pdf version of the documentation.
The other is to test for errors in the latex sections of the documentation.
(MathJax displays latex errors in red, but one has to check
every page that has latex to find all the errors this way.)
Once you have built *project_name*\ ``.tex``, the following command
executed in :ref:`config_file@directory@project_directory`
will accomplish both purposes:

   make -C *tex_directory* *project_name*\ ``.pdf``

#. The :ref:`config_file@project_name` is specified in the configuration file.
#. The resulting output file will be *project*\ ``.pdf`` in the
   *tex_directory* .
#. If a Latex error is encountered, the pdf build will stop with a message
   at the ``?`` prompt. If you enter ``q`` at this prompt, it will complete
   its processing in batch mode. You will be able to find the error messages
   in the file *project_name*\ ``.log`` in the *tex_directory* .
#. Translating Latex errors to the corresponding xrst input file:

   #. Latex error messages are reported using line numbers in
      the file *project*\ ``.tex`` .
   #. You may be able to find the corresponding xrst input file
      using by using ``grep`` to find text that is near the error.
   #. The page numbers in the :ref:`xrst_table_of_contents-title` are
      present in the latex input (often near ``section*{`` above the error)
      and may help translate these line numbers to page names.
   #. Given a page name, the corresponding xrst input file can
      be found at the top of the html version of the page.
{xrst_comment --------------------------------------------------------------- }

group_list
**********
It is possible to select one or more groups of pages
to include in the output using this optional argument.

#. The *group_list* is a list of one or more
   :ref:`group names<begin_cmd@group_name>`.
#. The :ref:`begin_cmd@group_name@Default Group` is represented by
   the group name ``default`` .
#. The order of the group names determines their order in the resulting output.
#. The default value for *group_list* is ``default`` .

For each group name in the *group_list*
there must be an entry in :ref:`config_file@root_file` specifying the
root file for that group name.

The xrst examples are a subset of its user documentation
and its user documentation is a subset of its developer documentation.
For each command, the same source code file provides both the
user and developer documentation. In addition, the developer documentation
has links to the user documentation and the user documentation has links
to the examples.

Example
=======
The examples commands below assume you have cloned the
`xrst git repository <https://github.com/bradbell/xrst>`_
and it is your current working directory.

#. The xrst examples use the default group
   and their documentation can be built using

      ``xrst --group_list default``

#. The xrst user documentation uses the default and user groups
   and its documentation can be built using

      ``xrst --group_list default user``

#. The xrst developer documentation uses the default, user, and dev
   groups and its documentation can be built using

      ``xrst --group_list default user dev``
{xrst_comment --------------------------------------------------------------- }

rename_group
************
If this option is present on the command line,
the :ref:`begin_cmd@group_name` in a subset of the source code, is changed.
This option replaces the :ref:`run_xrst@group_list`
by the list whose only entry is *new_group_name* .
None of the output files are created when rename_group is present;
e.g., the \*.rst and \*.html files.

old_group_name
==============
is the old group name for the pages that will have their group name replaced.
Use ``default``, instead of the empty group name, for the
:ref:`begin_cmd@group_name@Default Group` .

new_group_name
==============
Only the pages below the :ref:`config_file@root_file`
for *new_group_name* are modified.
You can rename a subset of the old group by making the root file
for the new group different than the root file for the old group.
Each page in the old group, and below the root file for the new group,
will have its group name changed from *old_group_name* to *new_group_name*.
Use ``default``, instead of the empty group name, for the
:ref:`begin_cmd@group_name@Default Group` .
{xrst_comment --------------------------------------------------------------- }

replace_spell_commands
**********************
If this option is present on the command line, the source code
:ref:`spell commands<spell_cmd-name>` are replaced in such a way that the
there will be no spelling warnings during future processing by xrst.
This is useful when there are no spelling warnings before a change
to the :ref:`config_file@project_dictionary` or when there is an update
of the pyspellchecker_ package (which is used to do the spell checking).
If this option is present,
none of the output files are created; e.g., the \*.rst and \*.html files.

.. _pyspellchecker: https://pypi.org/project/pyspellchecker

{xrst_end run_xrst}
"""
# ---------------------------------------------------------------------------
# imports
# ---------------------------------------------------------------------------
import sys
import re
import os
import toml
import string
import shutil
import filecmp
import argparse
import subprocess
# ---------------------------------------------------------------------------
# system_exit
# Error messages in this file do not use xrst.system_exit because
# they do not have file names that are relative to the project_directory.
def system_exit(msg) :
   # assert False, msg
   sys.exit(msg)
# ---------------------------------------------------------------------------
# Execute a system command
#
# command:
# the command to execute
#
# warning:
# If a warning is printed, this flag is set True
#
# page_name2line_pair
# a mapping from page name to to a list of line number pairs.
# The first number is the pair is a line number in the rst file for this page.
# The secon number is a corresponding xrst input line number.
#
# page_name2file_in
# a mapping from page name to the correspoind xrst input file.
#
def system_command(
      command                    ,
      warning                    ,
      page_name2line_pair = None ,
      page_name2file_in   = None ,
) :
   assert type(command) == str
   #
   assert type(warning) == list
   assert len(warning) == 1
   assert type(warning[0]) == bool
   #
   if page_name2line_pair != None :
      assert type(page_name2line_pair) == dict
      assert type(page_name2file_in) == dict
   #
   # subprocess.run, stderr
   print(command)
   command = command.split(' ')
   result = subprocess.run(command, capture_output = True)
   stderr = result.stderr.decode('utf-8')
   ok     =  result.returncode == 0 and stderr == ''
   if ok :
      return
   if page_name2line_pair == None :
      message  = stderr
      if result.returncode == 0 :
         message  += 'Warning: see system command message above.\n'
         sys.stderr.write(message)
         warning[0] = True
         return
      message  += 'Error: see system command message above.\n'
      system_exit(message)
   #
   # pattern_error
   pattern_error = re.compile( r'.*/rst/([a-z0-9_.]+).rst:([0-9]+):' )
   #
   # message
   message = ''
   #
   # error
   stderr_list = stderr.split('\n')
   for error in stderr_list :
      #
      # m_rst_error
      m_rst_error = pattern_error.search(error)
      if m_rst_error == None :
            if error != '' :
               # This should not happen
               # breakpoint()
               pass
      else :
         #
         # page_name
         page_name = m_rst_error.group(1)
         if not page_name in page_name2line_pair :
            assert page_name == 'xrst_table_of_contents'
         else :
            #
            # rst_line
            rst_line  = int( m_rst_error.group(2) )
            #
            # file_in, line_pair
            file_in   = page_name2file_in[page_name]
            line_pair = page_name2line_pair[page_name]
            #
            # n_pair
            n_pair = len(line_pair)
            #
            # index
            index  = 0
            while index < n_pair and line_pair[index][0] < rst_line :
                  index += 1
            #
            # line_before, line_after
            if index == n_pair :
               line_before = str( line_pair[n_pair-1][1] )
               line_after  = '?'
            elif line_pair[index][0] == rst_line :
               line_before = line_pair[index][1]
               line_after  = line_pair[index][1]
            elif 0 == index :
               line_before = '?'
               line_after  = line_pair[index][1]
            else :
               line_before = line_pair[index-1][1]
               line_after  = line_pair[index][1]
            #
            # error
            msg   = error[ m_rst_error.end()  : ]
            if line_before == line_after :
               error = f'{file_in}:{line_before}:{msg}'
            else :
               error = f'{file_in}:{line_before}-{line_after}:{msg}'
      #
      # message
      message += '\n' + error
   #
   if result.returncode == 0 :
      message  += 'Warning: see system command message above.\n'
      sys.stderr.write(message)
      warning[0] = True
      return
   message  += 'Error: see system command message above.\n'
   system_exit(message)
# ---------------------------------------------------------------------------
def fix_latex(latex_dir, project_name) :
   assert type(latex_dir) == str
   assert type(project_name) == str
   #
   # file_name
   file_name = f'{latex_dir}/{project_name}.tex'
   #
   # file_data
   file_obj  = open(file_name, 'r')
   file_data = file_obj.read()
   file_obj.close()
   #
   # file_data
   pattern   = re.compile( r'\n\\section{' )
   file_data = pattern.sub( r'\n\\section*{', file_data)
   #
   # file_data
   pattern   = re.compile( r'\n\\subsection{' )
   file_data = pattern.sub( r'\n\\subsection*{', file_data)
   #
   # file_data
   pattern   = re.compile( r'\n\\subsubsection{' )
   file_data = pattern.sub( r'\n\\subsubsection*{', file_data)
   #
   # file_name
   file_obj  = open(file_name, 'w')
   file_obj.write(file_data)
   file_obj.close()
   #
   return
# ---------------------------------------------------------------------------
# sys.path
# used so that we can test before installing
if( os.getcwd().endswith('/xrst.git') ) :
   if( os.path.isdir('xrst') ) :
      sys.path.insert(0, os.getcwd() )
#
import xrst
#
# version
# The script that updates version numbers expects version at begining of line
# and to have the value surrounded by single quotes.
version = '2023.1.12'
#
def run_xrst() :
   #
   # execution_directory
   execution_directory = os.getcwd()
   #
   # parser
   parser = argparse.ArgumentParser(
      prog='xrst', description='extract Sphinx RST files'
   )
   parser.add_argument('--version', action='store_true',
      help='just print version of xrst'
   )
   parser.add_argument('--local_toc', action='store_true',
      help='add a local table of contents at the top of each page'
   )
   parser.add_argument('--page_source', action='store_true',
      help='add link to the xrst source code be included at top of each page'
   )
   parser.add_argument(
      '--index_page_name', metavar='index_page_name', default='xrst_root_doc',
      help='The file index.html will be a redirect to this page' + \
         '(default is root of documentation tree)'
   )
   parser.add_argument(
      '--config_file', metavar='config_file', default='xrst.toml',
      help='location of the xrst configuration file which is in toml format' + \
         '(default is .)'
   )
   parser.add_argument(
      '--html_theme', metavar='html_theme', default='furo',
      help='sphinx html_theme that is used to display web pages ' + \
         '(default is furo)',
   )
   parser.add_argument(
      '--group_list', nargs='+', default='default',
      metavar= 'group_name' ,
      help='list of group_names to include in this build (default is default)'
   )
   parser.add_argument(
      '--rename_group', nargs=2, default=None,
      metavar=('old_group_name', 'new_group_name'),
      help='list of group_names to include in this build (default is default)'
   )
   parser.add_argument(
      '--target', metavar='target', choices=['html', 'tex'], default='html',
      help='type of output files, choices are html or tex (default is html)'
   )
   parser.add_argument('--replace_spell_commands', action='store_true',
      help='replace the xrst spell commands in source code files'
   )
   parser.add_argument('--rst_line_numbers', action='store_true',
      help='report sphinx errors and warnings using rst file line numbers'
   )
   parser.add_argument('--rst_only', action='store_true',
      help='report sphinx errors and warnings using rst file line numbers'
   )
   #
   # arguments
   arguments = parser.parse_args()
   #
   if arguments.version :
      print(version)
      sys.exit(0)
   #
   # local_toc
   local_toc = arguments.local_toc
   #
   # page_source
   page_source = arguments.page_source
   #
   # index_page_name
   index_page_name = arguments.index_page_name
   #
   # config_file
   # can not use system_exit until os.getcwd() returns project_directory
   config_file = arguments.config_file
   if not os.path.isfile(config_file) :
      msg  = 'xsrst: Error\n'
      msg += f'config_file = {config_file}\n'
      if config_file[0] == '/' :
         msg += 'is not a file\n'
      else :
         msg += f'is not a file relative to the execution directory\n'
         msg += execution_directory
      system_exit(msg)
   #
   # conf_dict
   conf_dict  = xrst.get_conf_dict(config_file)
   #
   # config_dir
   index = config_file.rfind('/')
   if 0 <= index :
      config_dir = config_file[: index]
      os.chdir(config_dir)
   #
   # project_directory
   project_directory = conf_dict['directory']['project_directory']
   #
   # make project directory the current working directory
   os.chdir(project_directory)
   #
   # replace_spell_commands
   replace_spell_commands = arguments.replace_spell_commands
   #
   # rst_line_numbers
   rst_line_numbers = arguments.rst_line_numbers
   #
   # rst_only
   rst_only = arguments.rst_only
   #
   # html_theme
   html_theme = arguments.html_theme
   #
   # group_list
   group_list = arguments.group_list
   if type(group_list) == str :
      group_list = [ group_list ]
   #
   # rename_group, group_list
   rename_group = arguments.rename_group
   if rename_group != None :
      group_list = [ rename_group[1] ]
      #
      if rename_group[0] == '' :
         msg = 'xrst rename_group: old_group_name is empty.\n'
         msg += 'Use "default" for the empty group name.'
         system_exit(msg)
      if rename_group[1] == '' :
         msg = 'xrst rename_group: new_group_name is empty.\n'
         msg += 'Use "default" for the empty group name.'
         system_exit(msg)
   #
   # replace_spell_commands or rename_group
   if replace_spell_commands or rename_group != None :
      if replace_spell_commands :
         option = 'replace_spell_commands'
      else :
         option = 'rename_group'
      #
      cwd      = os.getcwd()
      prompt   = f'\nThe {option} option will change some of \n'
      prompt  += 'the files read by xrst. Make sure that you have a backup\n'
      prompt  += f'of source files in {cwd}\n'
      prompt  += 'before contining this operation: continue [yes/no] ? '
      response = None
      while response not in [ 'yes', 'no' ]:
         response = input(prompt)
      if response != 'yes' :
         system_exit( f'xrst: aborting {option}' )
   #
   # target
   target = arguments.target
   #
   # target_directory
   if target == 'html' :
      target_directory = conf_dict['directory']['html_directory']
   else :
      assert target == 'tex'
      target_directory = conf_dict['directory']['tex_directory']
   #
   # rst_directory
   rst_directory = conf_dict['directory']['rst_directory']
   if rst_directory[0] == '/' :
      msg  = 'rst_directory = ' + rst_directory + '\n'
      msg += 'must be a path relative to the project_directory'
      xrst.system_exit(msg)
   #
   if not os.path.isdir(rst_directory) :
      os.makedirs(rst_directory)
   #
   # rst2project_directory
   # relative path from the rst_directory to the project directory
   rst2project_directory = os.path.relpath(
      os.getcwd() , rst_directory
   )
   #
   # tmp_dir
   tmp_dir = rst_directory + '/tmp'
   if os.path.isdir(tmp_dir) :
      shutil.rmtree(tmp_dir)
   os.mkdir(tmp_dir)
   #
   # _sources
   if not os.path.isdir(rst_directory + '/_sources') :
      os.mkdir(rst_directory + '/_sources')
   #
   # spell_checker
   spell_list  = list()
   for entry in conf_dict['project_dictionary']['data'] :
      word_list = entry.split('\n')
      for word in word_list :
         word = word.strip(' \t')
         if len(word) > 0 :
            spell_list.append(word)
   package       = conf_dict['spell_package']['data']
   spell_checker = xrst.get_spell_checker(spell_list, package)
   #
   # not_in_index_list
   not_in_index_list = list()
   for entry in conf_dict['not_in_index']['data'] :
      pattern_list = entry.split('\n')
      for pattern in pattern_list :
         pattern = pattern.strip(' \t')
         try :
            not_in_index_list.append( re.compile(pattern) )
         except :
            msg  = f'not_in_index table in config_file = {config_file}\n'
            msg += f'The regular expression "{pattern}" would not compile'
            system_exit(msg)
   # -------------------------------------------------------------------------
   #
   # input_file_list
   input_file_list = None
   #
   # root_file
   root_file = conf_dict['root_file']
   #
   # project_name
   project_name = conf_dict['project_name']['data']
   #
   # all_page_info
   # This list accumulates over all the group names
   all_page_info  = list()
   #
   # root_page_list
   # Each group has a root secion (in root_file) at the top if its tree.
   root_page_list = list()
   #
   # page_name2line_pair, page_name2file_in
   # Each rst page name has a corresponding input file and mapping from
   # rst file line nubmers to input file line numbers.
   page_name2line_pair = dict()
   page_name2file_in   = dict()
   #
   # any_warning
   any_warning = [ False ]
   #
   # group_name
   for group_name in group_list :
      #
      # old_group_name, new_group_name
      if rename_group == None :
         old_group_name = group_name
         new_group_name = group_name
      else :
         old_group_name = rename_group[0]
         new_group_name = rename_group[1]
         assert new_group_name == group_name
      #
      if new_group_name not in root_file :
         msg  = f'The group name {new_group_name} is '
         if rename_group == None :
            msg += 'in --group_list\n'
         else :
            msg += 'is new_group_name in --rename_group\n'
         msg += 'but it is not a valid key for the root_file table of\n'
         msg += f'the configuration file {config_file}'
         xrst.system_exit(msg)
      #
      if not os.path.isfile( root_file[new_group_name] ) :
         file_name = root_file[new_group_name]
         msg  = f'The root_file for group_name {new_group_name} is not a file\n'
         msg += f'file name = {file_name}'
         xrst.system_exit(msg)
      #
      # finfo_stack, finfo_done
      # This information is by file, not page
      finfo_stack      = list()
      finfo_done       = list()
      finfo = {
         'file_in'        : root_file[new_group_name],
         'parent_file'    : None,
         'parent_page'    : None,
      }
      finfo_stack.append(finfo)
      #
      while 0 < len(finfo_stack) :
         #
         # finfo
         # pop first element of stack so that order in tex file and
         # table of contents is correct
         finfo  = finfo_stack.pop(0)
         #
         for finfo_tmp in finfo_done :
            if finfo_tmp['file_in'] == finfo['file_in'] :
               msg  = 'The file ' + finfo['file_in']
               msg += '\nis included twice with '
               msg += f'group_name = "{old_group_name}"\n'
               msg += 'Once in ' + finfo_tmp['parent_file'] + '\n'
               msg += 'and again in ' + finfo['parent_file'] + '\n'
               xrst.system_exit(msg)
         finfo_done.append(finfo)
         #
         file_in              = finfo['file_in']
         parent_file          = finfo['parent_file']
         parent_file_page  = finfo['parent_page']
         assert os.path.isfile(file_in)
         #
         # get xrst docuemntation in this file
         file_page_info = xrst.get_file_info(
            all_page_info,
            old_group_name,
            parent_file,
            file_in,
         )
         #
         # root_page_list
         if finfo['parent_file'] == None :
            assert file_in == root_file[new_group_name]
            if file_page_info[0]['is_parent'] :
               n_page = 1
            else :
               n_page = len( file_page_info )
            for i_page in range( n_page ) :
               page_name = file_page_info[i_page]['page_name']
               root_page_list.append(page_name)
         #
         # parent_page_file_in
         # index in all_page_info of parent page for this file
         parent_page_file_in = None
         if file_page_info[0]['is_parent'] :
            parent_page_file_in = len(all_page_info)
         #
         # add this files pages to all_page_info
         for i_page in range( len(file_page_info) ) :
            # ------------------------------------------------------------
            # page_name, page_data, is_parent, begin_line
            page_name  = file_page_info[i_page]['page_name']
            page_data  = file_page_info[i_page]['page_data']
            is_parent  = file_page_info[i_page]['is_parent']
            is_child   = file_page_info[i_page]['is_child']
            begin_line = file_page_info[i_page]['begin_line']
            end_line   = file_page_info[i_page]['end_line']
            #
            # parent_page
            if is_parent or parent_page_file_in is None :
               parent_page = parent_file_page
            else :
               parent_page = parent_page_file_in
            #
            # all_page_info
            all_page_info.append( {
               'page_name'      : page_name,
               'file_in'        : file_in,
               'parent_page'    : parent_page,
               'in_parent_file' : is_child,
               'begin_line'     : begin_line,
               'end_line'       : end_line,
            } )
            # -------------------------------------------------------------
            # comment_command
            page_data = xrst.comment_command(page_data)
            # -------------------------------------------------------------
            # spell_command
            # do after suspend and before other commands to help ignore
            # pages of text that do not need spell checking
            page_data, spell_warning = xrst.spell_command(
               tmp_dir,
               page_data,
               file_in,
               page_name,
               begin_line,
               spell_checker,
            )
            if spell_warning :
               any_warning[0] = True
            # -------------------------------------------------------------
            # dir commands
            page_data = xrst.dir_command(page_data, rst2project_directory)
            # -------------------------------------------------------------
            # toc commands
            page_data, child_file, child_page_list = \
               xrst.toc_commands(
                  page_data,
                  file_in,
                  page_name,
                  old_group_name,
            )
            #
            # page_index, finfo_stack
            page_index = len(all_page_info) - 1
            for file_tmp in child_file :
               finfo_stack.append( {
                  'file_in'        : file_tmp,
                  'parent_file'    : file_in,
                  'parent_page'    : page_index,
               } )
            # ------------------------------------------------------------
            # code commands
            page_data = xrst.code_command(
               page_data,
               file_in,
               page_name,
               rst2project_directory,
            )
            # ------------------------------------------------------------
            # literal command
            page_data = xrst.literal_command(
               page_data,
               file_in,
               page_name,
               rst2project_directory,
            )
            # ------------------------------------------------------------
            # process headings
            #
            # pseudo_heading, all_page_info
            page_data, page_title, pseudo_heading = \
            xrst.process_headings(
               local_toc,
               page_data,
               file_in,
               page_name,
               not_in_index_list,
            )
            # all_page_info
            # page title is used by table_of_contents
            all_page_info[page_index]['page_title'] = page_title
            # -------------------------------------------------------------
            # list_children
            # page_name for each of the children of the current page
            list_children = child_page_list
            if is_parent :
               for i in range( len(file_page_info) ) :
                  if i != i_page :
                     list_children.append(
                        file_page_info[i]['page_name']
                     )
            # -------------------------------------------------------------
            # process children
            # want this as late as possible so toctree at end of input
            page_data = xrst.process_children(
               page_name,
               page_data,
               list_children,
            )
            # -------------------------------------------------------------
            #
            # line_pair and file tmp_dir/page_name.rst
            line_pair = xrst.temporary_file(
               page_source,
               target,
               pseudo_heading,
               file_in,
               tmp_dir,
               page_name,
               page_data,
            )
            #
            # page_name2line_pair, page_name2file_in
            page_name2line_pair[page_name] = line_pair
            page_name2file_in[page_name]   = file_in
      #
      # check_input_files
      if rename_group == None :
         toc_file_set = set()
         for finfo_tmp in finfo_done :
            toc_file_set.add( finfo_tmp['file_in'] )
         input_file_list, file_list_warning = xrst.check_input_files(
            config_file, conf_dict, group_name, toc_file_set, input_file_list
         )
         if file_list_warning :
            any_warning[0] = True
   #
   # index_page_name
   ok = index_page_name == 'xrst_root_doc' or target == 'tex'
   for page_info in all_page_info :
      ok = ok or page_info['page_name'] == index_page_name
   if not ok :
      msg = f'index_page_name = {index_page_name} is not a valid page name.'
      xrst.system_exit(msg)
   #
   # rename_group
   if rename_group != None :
      xrst.rename_group(tmp_dir, rename_group[0], rename_group[1])
      #
      # tmp_dir
      # reset tmp_dir because rmtree is such a dangerous command
      tmp_dir = f'{rst_directory}/tmp'
      shutil.rmtree(tmp_dir)
      print('xrst --rename_group: OK')
      return
   #
   # replace_spell_commands
   if replace_spell_commands :
      xrst.replace_spell(tmp_dir)
      #
      # tmp_dir
      # reset tmp_dir because rmtree is such a dangerous command
      tmp_dir = f'{rst_directory}/tmp'
      shutil.rmtree(tmp_dir)
      print('xrst --replace_spell_commands: OK')
      return
   #
   # auto_file
   xrst.auto_file(
      conf_dict, html_theme, target, all_page_info, root_page_list
   )
   #
   # -------------------------------------------------------------------------
   #
   # rst_directory/*.rst
   tmp_list = os.listdir(tmp_dir)
   rst_list = os.listdir(rst_directory)
   for name in tmp_list :
      src = f'{tmp_dir}/{name}'
      des = f'{rst_directory}/{name}'
      if name.endswith('.rst') :
         if name not in rst_list :
               shutil.copyfile(src, des)
         else :
            if not filecmp.cmp(src, des, shallow=False) :
               os.replace(src, des)
   for name in rst_list :
      if name.endswith('.rst') :
         if name not in tmp_list :
            os.remove( f'{rst_directory}/{name}' )
   #
   # tmp_dir
   # reset tmp_dir because rmtree is such a dangerous command
   tmp_dir = f'{rst_directory}/tmp'
   shutil.rmtree(tmp_dir)
   # -------------------------------------------------------------------------
   if rst_only :
      print('xrst rst_only: OK')
      indent = '\n' + 3 * ' '
      if target == 'html' :
         txt  = f'The following commands will build the html:'
         txt += indent
         txt += f'sphinx-build -b html {rst_directory} {target_directory}'
         txt += indent
         txt += f'rm -r {target_directory}/_sources'
         txt += indent
         txt += f'mv {rst_directory}/_sources {target_directory}/_sources'
         if html_theme == 'sphinx_rtd_theme' :
            txt += '\nThis will not modify the sphinx_rtd_theme maximum width'
      else :
         assert target == 'tex'
         latex_dir = f'{target_directory}'
         txt  = f'The following command will build the tex:'
         txt += indent
         txt += f'sphinx-build -b latex {rst_directory} {latex_dir}'
         txt += '\n'
         txt += f'The following command will build the pdf from the tex:'
         txt += indent
         txt += f'make -C {project_directory}/{latex_dir} {project_name}.pdf'
      print(txt)
      return
   #
   # -------------------------------------------------------------------------
   if target == 'html' :
      command = f'sphinx-build -b html {rst_directory} {target_directory}'
      if rst_line_numbers :
         system_command(command, any_warning)
      else :
         system_command(
            command, any_warning, page_name2line_pair, page_name2file_in
         )
         #
         # target_directory/_sources
         # replace sphinx _sources directory with proper xrst sources
         if not rst_line_numbers :
            src_dir = f'{rst_directory}/_sources'
            des_dir = f'{target_directory}/_sources'
            print( f'rm -r {des_dir}' )
            shutil.rmtree(des_dir)
            print( f'cp -r {src_dir} {des_dir}' )
            shutil.copytree(src_dir, des_dir)
         #
         # target_directory/index.html
         src_file = f'{target_directory}/{index_page_name}.html'
         des_file = f'{target_directory}/index.html'
         shutil.copyfile(src_file, des_file)

   else :
      assert target == 'tex'
      #
      latex_dir = f'{target_directory}'
      command = f'sphinx-build -b latex {rst_directory} {latex_dir}'
      if rst_line_numbers :
         system_command(command, any_warning)
      else :
         system_command(
            command, any_warning, page_name2line_pair, page_name2file_in
         )
      #
      # latex_dir/project_name.tex
      fix_latex(latex_dir, project_name)
      #
      print('The following command will build the pdf from the latex:')
      print( f'   make -C {project_directory}/{latex_dir} {project_name}.pdf' )
   # -------------------------------------------------------------------------
   # target_directory/_static/css/theme.css
   # see https://stackoverflow.com/questions/23211695/
   #  modifying-content-width-of-the-sphinx-theme-read-the-docs
   if html_theme == 'sphinx_rtd_theme' and target == 'html' :
      pattern = dict()
      pattern['content'] = re.compile(
         r'([.]wy-nav-content[{][^}]*;max-width):[^;]*;'
      )
      pattern['sidebar'] = re.compile(
         r'([.]wy-nav-side[{][^}]*;width):[^;]*;'
      )
      pattern['search'] = re.compile(
         r'([.]wy-side-nav-search[{][^}]*;width):[^;]*;'
      )
      new_value = { 'content':'100%', 'sidebar':'250px', 'search':'250px' }
      file_name = f'{target_directory}/_static/css/theme.css'
      if os.path.exists(file_name + '.bak' ) :
        shutil.copyfile(file_name + '.bak', file_name)
      else :
        shutil.copyfile(file_name, file_name + '.bak')
      try :
         file_obj  = open(file_name, 'r')
         ok        = True
      except :
         ok        = False
      if ok :
         data_in   = file_obj.read()
         file_obj.close()
         match     = dict()
         for key in pattern :
            match[key] = pattern[key].search(data_in)
            ok         = match[key] != None
      if ok :
         for key in pattern :
            match[key] = pattern[key].search(data_in, match[key].end())
            ok        = match[key] == None
      if ok :
         data_out = data_in
         for key in pattern :
            data_tmp  = data_out
            value     = new_value[key]
            data_out  = pattern[key].sub( f'\\1:{value};', data_tmp)
            ok        = data_out != data_tmp
      if not ok :
         msg       = 'warning: cannot modify widths in sphinx_rtd_theme\n'
         sys.stderr.write(msg)
         any_warning[0] = True
      else :
         file_obj  = open(file_name, 'w')
         file_obj.write(data_out)
         file_obj.close()
   # -------------------------------------------------------------------------
   if any_warning[0] :
      system_exit('xrst: See warnings above')
   else :
      print('xrst: OK')



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin auto_file user}
{xrst_spell
   conf
   config
}

Automatically Generated Files
#############################
These files are located in the :ref:`config_file@directory@rst_directory` .
A new version of these files is created each time ``xrst`` is run.
The files in the ``rst`` subdirectory that do not change are not replaced.
(This reduces the amount of processing that sphinx needs to do.)

conf.py
*******
This is the sphinx configuration_ file.
It is the only file in the *rst_directory* directory that is
not an rst file.

.. _configuration:  http://www.sphinx-doc.org/en/master/config


xrst_root_doc.rst
*****************
This file is the top of the documentation tree
built by sphinx. It is one level above the first page in
:ref:`config_file@root_file`.

xrst_table_of_contents.rst
**************************
This file contains the table of contents for the last run of ``xrst``.
You can link to the corresponding page with the following command::

   :ref:`xrst_table_of_contents-title`

The result of this command in this documentation is
:ref:`xrst_table_of_contents-title` .

xrst_index.rst
**************
If :ref:`run_xrst@target` is html,
this file contains a link to the table of contents generated by sphinx.
To be specific:

   :ref:`link to index<genindex>`

The xrst_index.rst file is not created when target is tex.

{xrst_end auto_file}
"""
# ----------------------------------------------------------------------------
import re
import os
import toml
import xrst
#
# conf_py_general
conf_py_general = '''#
# General configuration
extensions = [
   'sphinx.ext.mathjax',
]
root_doc = 'xrst_root_doc'
'''
#
# ----------------------------------------------------------------------------
# {xrst_begin auto_file_dev dev}
# {xrst_spell
#     bool
#     conf
#     dir
#     genindex
#     tmp
# }
# {xrst_comment_ch #}
#
# Create the automatically generated files
# ########################################
#
# conf_dict
# *********
# is a python dictionary representation of the configuration file.
#
# rst_dir
# =======
# we use *rst_dir* to denote *conf_dict* ['directory']['rst_directory'] .
#
# tmp_dir
# =======
# we use *tmp_dir* to denote *rst_dir*\ /tmp .
# This is the directory where xrst creates a temporary copy of *rst_dir* .
# These files are also automatically generated.
#
# html_theme
# **********
# The html_theme as on the xrst command line.
#
# target
# ******
# is html or tex
#
# all_page_info
# *************
# is a list with length equal to the number of pages.
# with the following key, value pairs (all the keys are strings):
#
# .. csv-table::
#     :header: key, value
#
#     page_name, (str) containing the name of this page.
#     page_title,  (str) containing the title for this page.
#     parent_page, (int) index in all_page_info for the parent of this page.
#     in_parent_file, (bool) is this page in same input file as its parent.
#
# root_page_list
# **************
# is a list of the root page names (one for each group) in the order
# they will appear in the table of contents.
#
# tmp_dir/xrst_table_of_contents.rst
# **********************************
# This file creates is the table of contents for the documentation.
# It has the label xrst_table_of_contents which can be used to link
# to this page.
#
# tmp_dir/xrst_index.rst
# **********************
# This file just contains a link to the genindex.rst file.
# It is (is not) included if target is html (tex).
#
# tmp_dir/xrst_root_doc.rst
# *************************
# This is the root level in the sphinx documentation tree.
#
# rst_dir/_sources
# ****************
# The sub-directory is used to store the replacement
# for the _sources directory. This contains the xrst sources that were used
# to create the rst files that sphinx used as sources.
# This is (is not) included if target is html (tex).
# If target is html, this sub-directory must exist and should be empty,
# before calling auto_file.
#
# rst_dir/conf.py
# ***************
# This is the configuration file used by sphinx to build the documentation.
#
#
# {xrst_code py}
def auto_file(
   conf_dict, html_theme, target, all_page_info, root_page_list
   ) :
   assert type(conf_dict) == dict
   assert type(html_theme) == str
   assert type(target) == str
   assert type(all_page_info) == list
   assert type(root_page_list) == list
   # {xrst_code}
   # {xrst_end auto_file_dev}
   #
   # rst_dir
   rst_dir = conf_dict['directory']['rst_directory']
   #
   # tmp_dir
   tmp_dir = f'{rst_dir}/tmp'
   #
   # project_name
   project_name = conf_dict['project_name']['data']
   #
   # ------------------------------------------------------------------------
   # tmp_dir/xrst_table_of_contents.rst
   #
   # file_data
   file_data     = xrst.table_of_contents(
      tmp_dir, target, all_page_info, root_page_list
   )
   #
   # xrst_table_of_contents.rst
   file_out    = tmp_dir + '/' + 'xrst_table_of_contents.rst'
   file_obj    = open(file_out, 'w')
   file_obj.write(file_data)
   file_obj.close()
   #
   # ------------------------------------------------------------------------
   # tmp_dir/xrst_index.rst
   if target == 'html' :
      #
      # file_data
      file_data  = 'Index\n'
      file_data += '#####\n'
      file_data += ':ref:`link to index<genindex>`'
      #
      # xrst_index.rst
      file_out    = tmp_dir + '/' + 'xrst_index.rst'
      file_obj    = open(file_out, 'w')
      file_obj.write(file_data)
      file_obj.close()
   # ------------------------------------------------------------------------
   # rst_dir/_sources
   if target == 'html' :
      #
      # file_in, file_daa
      file_in   = None
      file_data = None
      for info in all_page_info :
         if info['file_in'] != file_in :
            #
            # file_in, file_data
            file_in    = info['file_in']
            file_obj   = open(file_in, 'r')
            file_data  = file_obj.read()
            file_obj.close()
            #
            # newline_list
            newline_list = xrst.newline_indices(file_data)
         #
         # begin_index
         begin_line  = info['begin_line']
         assert 1 <= begin_line
         if begin_line == 1 :
            begin_index = 0
         else :
            assert begin_line < len(newline_list)
            begin_index = newline_list[begin_line-2] + 1
         #
         # end_index
         end_line  = info['end_line']
         assert end_line - 1 <= len(newline_list)
         if end_line - 1 == len(newline_list) :
            end_index = len(file_data)
         else :
            end_index = newline_list[end_line-1] + 1
         #
         # page_data
         page_data  = f'lines {begin_line}-{end_line} of file: {file_in}\n\n'
         page_data += file_data[begin_index : end_index]
         #
         # file_out
         page_name = info['page_name']
         file_out  = rst_dir + '/_sources/' + page_name
         if not page_name.endswith('.rst') :
            file_out += '.rst'
         file_out += '.txt'
         file_obj = open(file_out, 'w')
         file_obj.write(page_data)
         file_obj.close()
   # ------------------------------------------------------------------------
   # rst_dir/conf.py
   #
   # html_theme_options
   if html_theme not in conf_dict['html_theme_options'] :
      html_theme_options = None
   else :
      html_theme_options = conf_dict['html_theme_options'][html_theme]
      html_theme_optiosn = str( html_theme_options )
      assert '"' not in html_theme_options
   #
   # conf_py
   conf_py  = '# This file was automatically generated by xrst\n'
   conf_py += '#\n'
   conf_py += '# Project information\n'
   conf_py += f"project = '{project_name}'\n"
   conf_py +=  "author = ''\n"
   conf_py += conf_py_general
   conf_py += f'html_theme = "{html_theme}"\n'
   if html_theme_options != None :
      conf_py += f'html_theme_options = {html_theme_options}\n'
   #
   # rst_epilog, rst_prolog, latex_macro
   rst_epilog  = conf_dict['include_all']['rst_epilog']
   rst_prolog  = conf_dict['include_all']['rst_prolog']
   latex_macro = conf_dict['include_all']['latex_macro']
   #
   # rst_prolog
   if target == 'html' and len(latex_macro) != 0 :
      if rst_prolog != '' :
         rst_prolog += '\n\n'
      rst_prolog += '.. rst-class:: hidden\n\n'
      for macro in latex_macro :
         rst_prolog += 3 * ' ' + f':math:`{macro}`\n'
   #
   # conf_py
   if rst_epilog != '' :
      conf_py += '#\n'
      conf_py += f"rst_epilog = r'''\n"
      conf_py += rst_epilog
      conf_py += "'''\n"
   if rst_prolog != '' :
      conf_py += '#\n'
      conf_py += f"rst_prolog = r'''\n"
      conf_py += rst_prolog
      conf_py += "'''\n"
   if len( latex_macro ) != 0 :
      conf_py += '#\n'
      conf_py += '# Latex used when sphinx builds tex\n'
      conf_py += 'latex_elements = {\n'
      conf_py += 3 * ' ' + "'preamble' : r'''\n"
      for macro in latex_macro :
         conf_py += 3 * ' ' + macro + '\n'
      conf_py +=  3 * ' ' +"'''\n"
      conf_py += '}\n'
   #
   # rst_dir/conf.py
   file_out    = rst_dir + '/conf.py'
   file_obj    = open(file_out, 'w')
   file_obj.write(conf_py)
   file_obj.close()
   # ------------------------------------------------------------------------
   # tmp_dir/xrst_root_doc.rst
   #
   # file_data
   file_data  = '.. comment This file was automatically generated by xrst\n\n'
   file_data += project_name + '\n'
   file_data += len(project_name) * '#'
   file_data += '\n\n'
   file_data += '.. toctree::\n'
   file_data += '   :maxdepth: 1\n'
   file_data += '\n'
   file_data += '   xrst_table_of_contents\n'
   for page_name in root_page_list :
      file_data += '   ' + page_name + '\n'
   if target == 'html' :
      file_data += '   xrst_index\n'
   #
   # xrst_root_doc.rst
   file_name    = tmp_dir + '/xrst_root_doc.rst'
   file_obj     = open(file_name, 'w')
   file_obj.write(file_data)
   file_obj.close()



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
# {xrst_begin newline_indices dev}
# {xrst_comment_ch #}
#
# Find index of all the newlines in a string
# ##########################################
#
# Arguments
# *********
#
# data
# ====
# The string we are searching for newlines.
#
# Results
# *******
#
# newline_list
# ============
# The return newline_list is the list of indices in data that
# represent all of the newlines; i.e. '\n'.
#
# {xrst_code py}
def newline_indices(data) :
   assert type(data) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end newline_indices}

   pattern_newline  = re.compile( r'\n')
   newline_itr      = pattern_newline.finditer(data)
   newline_list     = list()
   for m_obj in newline_itr :
      next_index = m_obj.start()
      newline_list.append( next_index )
   # BEGIN_return
   assert type(newline_list) == list
   if 0 < len( newline_list) :
      assert type(newline_list[0]) == int
   return newline_list
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# -----------------------------------------------------------------------------
import re
import sys
import subprocess
# -----------------------------------------------------------------------------
# {xrst_begin check_input_files dev}
# {xrst_spell
#     conf
#     config
#     fullmatch
#     len
#     toc
# }
# {xrst_comment_ch #}
#
# Check That Expected xrst Input Files Are Included
# #################################################
#
# config_file
# ***********
# is the name of the configuration file.
#
# conf_dict
# *********
# is a python dictionary representation of the configuration file.
#
# group_name
# **********
# is the name of the group that we are checking
#
# toc_file_set
# ************
# is the set of files that were included by toc commands starting
# at the root file for this group.
# A warning is printed if a file has a begin command for this group
# and it is not in *toc_file_set*.
#
# file_list_in
# ************
# If file_list_in is None, the :ref:`config_file@input_files` commands
# will be executed to determine the file list.
# Otherwise, *file_list_in* will be used as the output of the first
# successful command.
#
# file_list_out
# *************
# This is a value that can be used for *file_list_in* to avoid
# having to re-execute the input_files commands.
#
# file_list_warning
# *****************
# This is true (false) if an input file list warning is (is not) printed
#
# Prototype
# *********
# {xrst_code py}
# file_list_out, file_list_warning =
def check_input_files(
   config_file, conf_dict, group_name, toc_file_set, file_list_in
) :
   #
   assert type(config_file) == str
   assert type(conf_dict) == dict
   assert type(group_name) == str
   assert type(toc_file_set) == set
   if file_list_in != None :
      assert type(file_list_in) == list
      if len(file_list_in) > 0 :
         assert type( file_list_in[0] ) == str
   #
   assert group_name != ''
   p_group_name = re.compile( r'[a-z]+' )
   assert p_group_name.fullmatch( group_name )
   # {xrst_code}
   # {xrst_literal
   #     # BEGIN_RETURN
   #     # END_RETURN
   # }
   # {xrst_end check_input_files}
   #
   # input_files
   input_files = conf_dict['input_files']['data']
   #
   # file_list_out
   if file_list_in != None :
      file_list_out = file_list_in
   elif len(input_files) == 0 :
      file_list_out = list()
   else :
      file_list_out = None
      command       = None
      for cmd in input_files :
         try :
            result  = subprocess.run(
               cmd,
               stdout   = subprocess.PIPE,
               stderr   = subprocess.PIPE,
               encoding = 'utf-8',
            )
         except:
            result = None
         if result == None :
            pass
         elif result.returncode == 0 :
            command       = cmd
            file_list_out = result.stdout
            if 0 <= file_list_out.find('"')  or 0 <= file_list_out.find("'") :
               msg  = 'warning: single or double quote in output of '
               msg += f'the input_files command: {command}\n'
               msg += 'This feature is not yet supported\n'
               sys.write(msg)
               file_list_out == list()
            else :
               file_list_out = file_list_out.split()
      if file_list_out != None :
         command = ' '.join(command)
         msg = f'Using following input_files: {command}\n'
         sys.stdout.write(msg)
      elif len(input_files) > 0 :
         msg  = 'warning: None of the commands in '
         msg += f'{config_file}.input_files succeeded\n'
         sys.stderr.write(msg)
         file_list_out = list()
   #
   # p_empty
   p_empty  = r'(^|[^\\])\{xrst_(begin|begin_parent)[ \t]+([^ \t}]*)[ \t]*}'
   p_empty  = re.compile(p_empty)
   #
   # p_non_empty
   p_non_empty  = r'(^|[^\\])\{xrst_(begin|begin_parent)[ \t]+([^ \t}]*)[ \t]+'
   p_non_empty += group_name
   p_non_empty += r'[ \t]*}'
   p_non_empty  = re.compile( p_non_empty )
   #
   # group_file_set
   group_file_set = set()
   #
   # warning_count
   warning_count = 0
   #
   # file_name
   for file_name in file_list_out :
      if warning_count < 10 :
         #
         # file_data
         try :
            file_obj     = open(file_name, 'r')
            file_data    = file_obj.read()
            file_obj.close()
         except :
            file_data = ''
         #
         # m_non_empty
         m_non_empty = p_non_empty.search( file_data )
         if m_non_empty != None :
            if file_name not in toc_file_set :
               if warning_count == 0 :
                  msg  = '\nwarning: group_name = ' + group_name + '\n'
                  msg += 'The following files have pages with this group name '
                  msg += 'but they are not in any xrst_toc commands '
                  msg += 'starting at the root_file for this group\n'
                  sys.stderr.write(msg)
               msg = 3 * ' ' +file_name + '\n'
               sys.stderr.write(msg)
               warning_count += 1
         #
         # m_empty
         elif group_name == 'default' :
            m_empty = p_empty.search( file_data )
            if m_empty != None :
               if file_name not in toc_file_set :
                  if warning_count == 0 :
                     msg  = '\nwarning: group_name = ' + group_name + '\n'
                     msg += 'The following files have pages with '
                     msg += 'the empty group name\n'
                     msg += 'but they are not in any xrst_toc commands '
                     msg += 'starting at the root_file for the default group\n'
                     sys.stderr.write(msg)
                  msg = 3 * ' ' +file_name + '\n'
                  sys.stderr.write(msg)
                  warning_count += 1
            #
            if warning_count == 10 :
               msg+= f'Surpressing this warning after {warning_count} files.\n'
               sys.stderr.write(msg)
   file_list_warning = 0 < warning_count
   # BEGIN_RETURN
   assert type(file_list_warning) == bool
   assert type(file_list_out) == list
   if len(file_list_out) > 0 :
      assert type( file_list_out[0] ) == str
   return file_list_out, file_list_warning
   # END_RETURN



# -----------------------------------------------------------------------------
#                      xrst: Extract Sphinx RST Files
#          Copyright (C) 2020-22 Bradley M. Bell (bradbell@seanet.com)
#              This program is distributed under the terms of the
#              GNU General Public License version 3.0 or later see
#                    https://www.gnu.org/licenses/gpl-3.0.txt
# ----------------------------------------------------------------------------
import re
# {xrst_begin xrst.pattern dev}
# {xrst_spell
#     arg
#     dir
#     lin
#     toc
# }
# {xrst_comment_ch #}
#
# The xrst.pattern Dictionary
# ###########################
#
# pattern
# *******
# This dictionary contains compiled regular expressions.
# It does not change after its initial setting when this file is imported.
# {xrst_code py}
pattern = dict()
# {xrst_code}
#
# begin
# *****
# Pattern for the begin command.
#
# 0. preceding character or empty + the command.
# 1. preceding character or empty
# 2. begin or begin_parent
# 3. the page name (without leading or trailing spaces or tabs)
# 4. the group name (with leading and trailing spaces and tabs)
#
# {xrst_code py}
pattern['begin'] = re.compile(
   r'(^|[^\\])\{xrst_(begin|begin_parent)[ \t]+([^ \t}]*)([^}]*)}'
)
# {xrst_code}
#
# toc
# ***
# Patterns for the toc_hidden, toc_list, and toc_table commands.
#
# 0. preceding character + the command.
# 1. command name; i.e., hidden, list, or table
# 2. the rest of the command that comes after the command name.
#    This is a list of file names with one name per line.
#    The } at the end of the command is not included.
#    This pattern may be empty.
#
# If you change this pattern, check pattern_toc in process_children.py
# {xrst_code py}
pattern['toc']   = re.compile(
   r'[^\\]\{xrst_toc_(hidden|list|table)([^}]*)}'
)
# {xrst_code}
#
# code
# ****
# Pattern for code command.
#
# 0. the entire line for the command with newline at front.
# 1. the indent for the command (spaces and tabs)
# 2. is the command with or without characters in front
# 3. This is the non space characters after the indent and before
#    command (or None)
# 4. the language argument which is empty (just white space)
#    for the second code command in each pair.
# 5. the line number for this line; see pattern['line'] above.
#
# {xrst_code py}
pattern['code'] = re.compile(
   r'\n([ \t]*)(\{xrst_code *|([^\n]*[^\n\\])\{xrst_code *)' +
   r'([^}]*)}[^\n]*\{xrst_line ([0-9]+)@'
)
# {xrst_code}
#
# comment_ch
# **********
# Pattern for comment_ch command
#
# 1. empty or character before command + the command
# 2. is the character (matched as any number of not space, tab or }
#
# {xrst_code py}
pattern['comment_ch'] = re.compile(
   r'(^|[^\\])\{xrst_comment_ch\s+([^} \t]*)\s*}'
)
# {xrst_code}
#
# dir
# ***
# Pattern for dir command
#
# 1. Is either empty of character before command
# 2. Is the file_name in the command
#
# {xrst_code py}
# pattern_dir
pattern['dir'] = re.compile(
   r'(^|[^\\]){xrst_dir[ \t]+([^}]*)}'
)
# {xrst_code}
#
# end
# ***
# Pattern for end command
#
# 0. preceding character + white space + the command.
# 1. the page name.
#
# {xrst_code py}
pattern['end'] = re.compile( r'[^\\]\{xrst_end\s+([^}]*)}' )
# {xrst_code}
#
#
# line
# ****
# Pattern for line numbers are added to the input by add_line_number
#
# 0. the line command.
# 1. the line_number.
#
# {xrst_code py}
pattern['line'] = re.compile( r'\{xrst_line ([0-9]+)@' )
# {xrst_code}
#
# arg, lin
# ********
#
# literal_0
# *********
# xrst_literal with no arguments
#
# 0. preceding newline + white space + the command.
# 1. line number where } at end of command appears
#
# {xrst_code py}
lin = r'[ \t]*\{xrst_line ([0-9]+)@\n'
pattern['literal_0'] = re.compile(
   r'[^\\]\{xrst_literal}' + lin
)
# {xrst_code}
#
# literal_1
# *********
# xrst_literal with display_file
#
# 0. preceding newline + white space + the command.
# 1. the line number where this command starts
# 2. the display file
# 3. line number where display file appears
# 4. line number where } at end of command appears
#
# {xrst_code py}
arg = r'([^{]*)\{xrst_line ([0-9]+)@\n'
pattern['literal_1']  = re.compile(
   r'[^\\]\{xrst_literal' + lin + arg + r'[ \t]*}' + lin
)
# {xrst_code}
#
# literal_2
# *********
# xrst_literal with start, stop
#
# 0. preceding newline + white space + the command.
# 1. the line number where this command starts
# 2. the start text + surrounding white space
# 3. line number where start text appears
# 4. the stop text + surrounding white space
# 5. the line number where stop text appears
# 6. line number where } at end of command appears
#
# {xrst_code py}
pattern['literal_2']  = re.compile(
   r'[^\\]\{xrst_literal' + lin + arg + arg + r'[ \t]*}' + lin
)
# {xrst_code}
#
# literal_3
# *********
# xrst_literal with start, stop, display_file
#
# 0. preceding character + the command.
# 1. the line number where this command starts
# 2. the display file
# 3. line number where display file appears
# 4. the start text + surrounding white space
# 5. line number where start text appears
# 6. the stop text + surrounding white space
# 7. the line number where stop text appears
# 8. line number where } at end of command appears
#
# {xrst_code py}
pattern['literal_3']  = re.compile(
   r'[^\\]\{xrst_literal' + lin + arg + arg + arg + r'[ \t]*}' + lin
)
# {xrst_code}
# {xrst_end xrst.pattern}



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin comment_ch_cmd user}

Comment Character Command
#########################

Syntax
******
``\{xrst_comment_ch`` *ch* ``}``

Purpose
*******
Some languages have a special character that
indicates the rest of the line is a comment.
If you embed sphinx documentation in this type of comment,
you need to inform xrst of the special character so it does
not end up in your ``.rst`` output file.

ch
**
The value of *ch* must be one non white space character.
There must be at least one white space character
between ``xrst_comment_ch`` and *ch*.
Leading and trailing white space around *ch* is ignored.

Input Stream
************
Spaces and tabs before the special character,
the special character,
and one space directory after the special character (if present),
are removed from the input stream before any xrst processing.
For example, if :code:`#` is the special character,
the following input has the heading Factorial
and the ``def`` token indented the same amount:

.. code-block:: py

   # Factorial
   # ---------
   def factorial(n) :
      if n == 1 :
         return 1
      return n * factorial(n-1)

Indent
******
The :ref:`indent-name` is calculated before removing the special character
and its white space.

Example
*******
:ref:`comment_ch_example-name`

{xrst_end comment_ch_cmd}
"""
# ----------------------------------------------------------------------------
import re
import xrst
#
# pattern
pattern = dict()
# need \{xrst_comment_ch so it does not match comment_ch command
pattern['error']      = re.compile( r'[^\\]\{xrst_comment_ch[^a-z]' )
pattern['comment_ch'] = xrst.pattern['comment_ch']
#
# Returns the beginning of line comment character for a file.
#
# Comment Character:
# The comment character is specified by \{xrst_comment_ch ch} where ch
# is a single character after leading and trailing white space is removed.
#
# data_in:
# is the data for this page.
#
# data_out:
# is a copy of data_in with the comment command removed.
#
# file_name:
# is the name of this file (used for error reporting).
#
# page_name:
# is the page name (used for error reporting)
#
# comment_ch:
# is a the comment character for this file. It is None when the
# is no comment character command for the file.
#
# data_out, comment_ch =
def comment_ch_command(data_in, file_name, page_name) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   #
   # m_obj
   m_obj   = pattern['comment_ch'].search(data_in)
   if not m_obj :
      m_error = pattern['error'].search(data_in)
      if m_error :
         msg  = f'syntax_error in xrst comment_ch command'
         line = data_in[: m_error.start() + 1].count('\n') + 1
         xrst.system_exit(
            msg,
            file_name = file_name,
            page_name = page_name,
            m_obj     = m_error,
            data      = data_in,
         )
      #
      # data_out
      data_out = data_in
      #
      # comment_ch
      comment_ch = None
   else :
      #
      # comment_ch
      comment_ch = m_obj.group(2)
      line       = data_in[: m_obj.start() ].count('\n') + 1
      if len( comment_ch ) != 1 :
         msg = 'Expected a single character argument to comment_ch command'
         xrst.system_exit(
            msg,
            file_name = file_name,
            page_name = page_name,
            m_obj     = m_obj,
            data      = data_in,
         )
      if comment_ch in '.:]' :
         msg  = f'Cannot use {comment_ch} as character in comment_ch command\n'
         xrst.system_exit(msg, file_name=file_name, line=line)
      #
      # m_rest
      data_rest  = data_in[ m_obj.end() : ]
      m_rest     = pattern['error'].search(data_rest)
      if m_rest :
         line = data_in[: m_obj.end() + m_rest.start() ].count('\n') + 1
         msg = 'There are multiple comment_ch commands in this file'
         xrst.system_exit(
            msg,
            file_name = file_name,
            page_name = page_name,
            m_obj     = m_rest,
            data      = data_in,
         )
      #
      # data_out
      data_out= data_in[: m_obj.start()] + data_in[ m_obj.end() :]
      #
   #
   # check_syntax_error
   xrst.check_syntax_error(
      command_name  = 'comment_ch',
      data          = data_out,
      file_name     = file_name,
      page_name     = page_name,
   )
   #
   return data_out, comment_ch



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
# {xrst_begin add_before_title dev}
# {xrst_spell
#     overline
# }
# {xrst_comment_ch #}
#
# If PDF, Add Page Number and Name to Title
# #########################################
#
# Arguments
# *********
#
# data_in
# =======
# data for this page before replacement.
#
#  #. data_in must contain '\\n{xrst_before_title}'
#     which is referred to as the command below.
#  #. The page title must come directly after the command
#     and start with a newline.
#  #. The page title may have an rst overline directly before the
#     heading text and must have an underline directly after it.
#  #. If both an overline and underline follow, they must be equal.
#
# target
# ======
# if *target* is ``html`` , the command is removed and no other action
# is taken. Otherwise, the *page_number* following by the *page_name* is
# added at the font of the title for this page.
# The underline (and overline if present) are extended by the number of
# characters added to the title.
#
# page_number
# ===========
# This is a page number that identifies this page in the table of contents.
#
# page_name
# =========
# This is the name of the page.
#
# Returns
# *******
#
# data_out
# ========
# the return data_out is the data after replacement.
#
# {xrst_code py}
def add_before_title(data_in, target, page_number, page_name) :
   assert type(data_in) == str
   assert target == 'html' or target == 'tex'
   assert type(page_number) == str
   assert type(page_name) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end add_before_title}
   #
   # pattern
   pattern   = '\n{xrst_before_title}'
   #
   # punctuation
   # Headings uses repeated copies of one of these characters
   punctuation = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
   #
   # start_cmd
   start_cmd = data_in.find(pattern)
   assert 0 <= start_cmd
   index = data_in.find(pattern, start_cmd + len(pattern))
   if 0 <= index :
      msg = '{xrst_before_title} cannot appear at beginning of a line'
      xrst.system_exit(msg, page_name = page_name)
   #
   # data_out
   if target == 'html' :
      return data_in.replace(pattern,'')
   #
   # add_text
   if page_number == '' :
      add_text = f'{page_name}:'
   else :
      add_text = f'{page_number} {page_name}:'
   #
   # first_newline
   first_newline = start_cmd + len(pattern)
   assert data_in[first_newline] == '\n'
   #
   # second_newline
   second_newline = data_in.index('\n', first_newline + 1)
   #
   # third_newline
   third_newline  = data_in.index('\n', second_newline + 1)
   #
   # first_line, second_line
   first_line   = data_in[first_newline + 1 : second_newline ]
   second_line  = data_in[second_newline + 1 : third_newline ]
   #
   # overline
   # check for an overline directly after the command
   overline = False
   if first_line[0] * len(first_line) == first_line :
      if first_line[0] in punctuation :
         overline = True
   #
   if not overline :
      # If no overline the page title follows the command and then
      # an underline after the title.
      assert second_line[0] in punctuation
      #
      # new version of first and second lines
      first_line   = add_text + ' ' + first_line
      second_line += second_line[0] * ( len(add_text) + 1 )
      #
   else :
      # fourth_newline
      fourth_newline = data_in.index('\n', third_newline + 1)
      #
      # third_line
      third_line   = data_in[third_newline + 1 : fourth_newline]
      assert first_line == third_line
      #
      # new version of first, second, and third lines
      first_line += first_line[0] * ( len(add_text) + 1 )
      second_line = add_text + ' ' + second_line
      third_line  = first_line
   #
   # data_out
   data_out  = data_in[: start_cmd] + '\n'
   data_out += first_line + '\n'
   data_out += second_line + '\n'
   if overline :
      data_out += third_line + '\n'
      data_out += data_in[fourth_newline :]
   else :
      data_out += data_in[third_newline :]
   # BEGIN_return
   assert type(data_out) == str
   #
   return data_out
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
# {xrst_begin check_syntax_error dev}
# {xrst_comment_ch #}
#
#
# Check that an xrst command has been removed
# ###########################################
#
# command_name
# ************
# is the name of the xrst command; i.e., the following pattern
# constitute a match for this command
#
# - ``[^\\]\{xrst_`` *command_name* ``[^z-a]``
#
# data
# ****
# is the data for this page.
#
# file_name
# *********
# is the input that this page appears in (used for error reporting).
#
# page_name
# *********
# is ``None`` or the name of this page (used for error reporting).
#
# {xrst_code py}
def check_syntax_error(command_name, data, file_name, page_name) :
   assert type(command_name) == str
   assert type(data) == str
   assert type(file_name) == str
   assert type(page_name) == str or page_name == None
   # {xrst_code}
   # {xrst_end check_syntax_error}
   #
   pattern = r'[^\\]{xrst_' + command_name + r'[^a-z]'
   m_error = re.search(pattern, data)
   if m_error :
      msg = f'syntax error in xrst {command_name} command'
      xrst.system_exit(msg,
         file_name    = file_name ,
         page_name    = page_name ,
         m_obj        = m_error ,
         data         = data ,
      )
   #
   return



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
#
import sys
import re
import os
import xrst
#
# usage
usage = \
'''
usage: update_xrst.py operation file_in file_out

operation: is one of the following: dot2atsign
file_in:   is the name of the file we are updating.
file_out:  is the name of the updated file. It can be the samle as file_in.

comment_ch:
Change comment_ch command from file scope to page scope.

ref_section_2:
Change :ref:`section_name-0` -> :ref:`section_name-0`,

literal_order:
{xrst_literal start stop display_file}->{xrst_literal display_file start stop}

tab3space:
Change tabs to 3 spaces.

space4to3:
Change tab stop from 4 spaces to 3 spaces.

file2literal:
change xrst_file -> xrst_literal.

child2toc:
Change xrst_children ->, xrst_toc_hidden, xrst_child_list -> xrst_toc_list,
and xrst_child_table -> xrst_toc_table.

ref_section:
Change :ref:`section_name-name` -> :ref:`section_name-0`,
Change :ref:`section_name<section_name-name>` -> :ref:`section_name-name`

dot2atsign:
Change the '.' to '@' character in all text of the form :ref:`text-name` and
text of the form (at the betinning of a line) .. _text:

xsrst2xrst:
Change 'xsrst' to 'xrst' in all the text.

'''
#
#
# abort
def abort(msg) :
   msg = '\nupdate_xstst.py: ' + msg
   sys.exit(msg)
#
# ref_section_2
def ref_section_2(data_in) :
   #
   # pattern
   pattern = re.compile( r':ref:`@([._a-z0-9]+)`' )
   #
   # data_out
   data_out  = pattern.sub( r':ref:`\1-0`', data_in)
   return data_out
#
# comment_ch:
def comment_ch(data_in) :
   # data_out
   data_out = data_in
   #
   # command
   pattern = re.compile(
      r'(^|[^\\])\{xrst_comment_ch\s+([^} \t]*)\s*}'
   )
   m_obj = pattern.search(data_out)
   if m_obj == None :
      return data_out
   comment_ch = m_obj.group(2)
   command    = comment_ch + ' {xrst_comment_ch ' + comment_ch + '}\n'
   #
   # data_out
   start = m_obj.start() - 1
   while start > 0 and data_out[start] in '[ \t]' :
      start -= 1
   assert data_out[start] == comment_ch
   end   = m_obj.end()
   while data_out[end] in '[ \t]' :
      end += 1
   assert data_out[end] == '\n'
   data_out = data_out[: start] + data_out[end +1 :]
   #
   # pattern
   pattern = re.compile(
      r'([^\\]{xrst_(begin|begin_parent)[ \t]+[^}]*}[^\n]*\n)([ \t]*)'
   )
   #
   # m_begin
   m_begin = pattern.search(data_out)
   while m_begin :
      #
      # indent
      indent = m_begin.group(3)
      #
      # data_out
      end      = m_begin.end()
      data_out = data_out[: end] + command + indent + data_out[end :]
      #
      m_begin = pattern.search(data_in, end + len(command))
   return data_out
#
# literal_order:
def literal_order(data_in) :
   # data_out
   data_out = data_in
   #
   # pattern
   pattern = re.compile(
      r'{xrst_literal[ \t]*\n([^}\n]*)\n([^}\n]*)\n([^}\n]*)\n([^}]*)}'
   )
   #
   # m_obj
   m_obj = pattern.search(data_out)
   while m_obj :
      #
      # command
      command  = '{xrst_literal\n'
      command += m_obj.group(3) + '\n'
      command += m_obj.group(1) + '\n'
      command += m_obj.group(2) + '\n'
      command += m_obj.group(4) + '}'
      #
      # data_out, data_left
      data_before = data_out[:  m_obj.start()]
      data_left   = data_before + command
      data_after  = data_out[m_obj.end() :]
      data_out    = data_before + command + data_after
      #
      # m_obj
      m_obj = pattern.search(data_out, len(data_left) )
   #
   return data_out
#
# tab3space:
def tab3space(data_in) :
   # data_out
   data_out = data_in
   #
   # pattern
   # group(1): start of line
   # group(2): possible newline_ch character
   # group(3): one or more of tabs
   pattern  = re.compile( r'(^|\n)([^\n\t]?)(\t{1,})' )
   #
   # m_obj
   m_obj   = pattern.search(data_out)
   #
   # data_out
   while m_obj :
      #
      # n_tab
      n_tab = len( m_obj.group(3) )
      assert 0 < n_tab
      #
      # replace
      if m_obj.group(2) == '' :
         replace  = m_obj.group(1) + n_tab * (3 * ' ')
      else :
         replace  = m_obj.group(1) + m_obj.group(2) + 2 * ' '
         replace += (n_tab - 1) * (3 * ' ')
      #
      # data_left, data_out
      data_left  = data_out[: m_obj.start()] + replace
      data_right = data_out[m_obj.end() :]
      data_out   = data_left + data_right
      #
      # m_obj
      m_obj    = pattern.search(data_out, len(data_left) )
   #
   # pattern
   # group(1): one non newline or tab
   # group(2): one or more tabs
   pattern  = re.compile( r'([^\n\t])(\t{1,})' )
   #
   # m_obj
   m_obj = pattern.search(data_out)
   #
   # data_out
   while m_obj :
      #
      # n_tab
      n_tab = len( m_obj.group(2) )
      assert 0 < n_tab
      #
      # replace
      replace  = m_obj.group(1) + 2 * ' '
      replace += (n_tab - 1) * (3 * ' ')
      #
      #
      # data_left, data_out
      data_left  = data_out[: m_obj.start()] + replace
      data_right = data_out[m_obj.end() :]
      data_out   = data_left + data_right
      #
      # m_obj
      m_obj    = pattern.search(data_out, len(data_left) )
   #
   if 0 <= data_out.find('\t') :
      msg = 'update_xrst.py tab3space: not all tabs have been replaced'
      assert False, msg
   #
   return data_out
#
# space4to3:
def space4to3(data_in) :
   # data_out
   data_out = data_in
   #
   # data_out
   pattern  = re.compile( r'(^|\n)([0-9#.]\.)  ([^ ])' )
   data_out = pattern.sub(r'\1\2 \3', data_out)
   #
   # data_out
   pattern = re.compile( r'(^|\n)-   ([^ ])' )
   data_out = pattern.sub(r'\1-  \2', data_out)
   #
   # pattern
   pattern = re.compile( r'(^|\n)([# ]   (    ){0,})[^ ]' )
   #
   # m_obj
   m_obj   = pattern.search(data_out)
   while m_obj :
      #
      # start, end
      start = m_obj.start() + len( m_obj.group(1) )
      end   = m_obj.end() - 1
      assert end - start == len( m_obj.group(2) )
      #
      # n_indent
      n_indent = int ( (end - start) / 4 )
      assert (end - start) == n_indent * 4
      #
      # data_out
      replace  = n_indent * '   '
      if data_out[start] != ' ' :
         replace = data_out[start] + replace[1 :]
      data_out = data_out[: start] + replace + data_out[end :]
      #
      # m_obj
      m_obj    = pattern.search(data_out, start + n_indent * 3 )
   #
   # data_out
   pattern = re.compile( r'(^|\n)((   ){0,}){   ([^ ])' )
   data_out = pattern.sub( r'\1\2{  \4', data_out )
   #
   pattern  = re.compile( r'(\n# {8})(Copyright \(C\))' )
   data_out = pattern.sub( r'\1   \2', data_out )
   #
   pattern  = re.compile( r'(\n {3})(GNU Affero General Public)' )
   data_out = pattern.sub( r'\1 \2', data_out )
   # -----------------------------------------------------------------------
   return data_out
#
# fiile2literal:
def file2literal(data_in) :
   data_out = data_in.replace('{xrst_file', '{xrst_literal')
   return data_out
#
# child2toc:
def child2toc(data_in) :
   data_out = data_in.replace('{xrst_children', '{xrst_toc_hidden')
   data_out = data_out.replace('{xrst_child_list', '{xrst_toc_list')
   data_out = data_out.replace('{xrst_child_table', '{xrst_toc_table')
   return data_out
#
# ref_section
def ref_section(data_in) :
   #
   # pattern
   pattern = dict()
   pattern['title']        = re.compile( r':ref:`([._a-z0-9]+)`' )
   pattern['section_name'] = re.compile( r':ref:`([._a-z0-9]+)<\1>`' )
   #
   # data_out
   data_out  = data_in
   data_out  = pattern['title'].sub( r':ref:`@\1`', data_out)
   data_out  = pattern['section_name'].sub( r':ref:`\1`', data_out)
   data_out  = data_out.replace(':ref:`genindex-0`', ':ref:`genindex`')
   return data_out
#
# dot2atsign
def dot2atsign(data_in) :
   # pattern
   pattern        = dict()
   pattern['ref'] = re.compile( r'(:ref:`[^`]*)\.' )
   #
   # data_out
   data_out = data_in
   #
   # m_ref
   m_ref    = pattern['ref'].search(data_out)
   while m_ref :
      # data_out
      data_out = pattern['ref'].sub(r'\1@', data_out)
      #
      # m_ref
      m_ref  = pattern['ref'].search(data_out, m_ref.start() )
   #
   return data_out
#
# xsrst2xrst
def xsrst2xrst(data_in) :
   data_out = data_in.replace('xsrst', 'xrst')
   return data_out
#
# main
def main() :
   if( len(sys.argv) != 4 ) :
      print(usage)
   #
   # operation_dict
   operation_dict = {
      'comment_ch'    :  comment_ch,
      'ref_section_2' :  ref_section_2,
      'literal_order' :  literal_order,
      'tab3space'     :  tab3space,
      'space4to3'     :  space4to3,
      'file2literal'  :  file2literal,
      'child2toc'     :  child2toc,
      'ref_section'   :  ref_section,
      'dot2atsign'    :  dot2atsign,
      'xsrst2xrst'    :  xsrst2xrst,
   }
   #
   # operation
   operation = sys.argv[1]
   if operation not in operation_dict :
      msg  = f'operation = {operation} is not a valid operation'
      abort(msg)
   #
   # file_in,
   file_in = sys.argv[2]
   if not os.path.isfile(file_in) :
      msg = f'file_in = {file_in} is not an existing file'
      abort(msg)
   #
   # file_out
   file_out = sys.argv[3]
   #
   # file_data
   file_obj   = open(file_in, 'r')
   file_data  = file_obj.read()
   file_obj.close()
   #
   # file_out
   file_data = operation_dict[operation](file_data)
   #
   # file_out
   file_obj = open(file_out, 'w')
   file_obj.write(file_data)
   file_obj.close()
#
main()
print('update_xrst.py: OK')
sys.exit(0)



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin heading_example}

Heading and Link Example
########################
Both the page name and page title are level zero headings for a page.
Using the *page_name* label displays the page name as its linking text;
e.g. for this page both the label and linking text are::

   heading_example

The *page_name*  name followed by ``-title`` displays the page title
as its linking text;
e.g. for this page the label is ``heading_example-title``
and linking text is::

   Heading and Link Example


Level One
*********
The label for this heading is ``heading_example@Level One``.

Level Two
=========
The label for this heading is ``heading_example@Level One@Level Two``.

Another Level One
*****************
The label for this heading is ``heading_example@Another Level One``.

Level Two\_
===========
The label for this heading is
``heading_example@Another Level One@Level Two_``.
Note that the backslash in the heading keeps ``Two_``
from being interpreted as a link.
Also note that the backslash does not appear in the
display of the heading or in the corresponding label.

x
=
A heading can have just one character.
The label for this heading is
``heading_example@Another Level One@x``.

Links
*****
These links would also work from any other page because the page name
(``heading_example`` in this case)
is included at the beginning of the target for the link:

#. :ref:`heading_example-name`
#. :ref:`heading_example-title`
#. :ref:`heading_example@Level One`
#. :ref:`heading_example@Level One@Level Two`
#. :ref:`heading_example@Another Level One`
#. :ref:`heading_example@Another Level One@Level Two_`
#. :ref:`heading_example@Another Level One@x`

Linking Headings Using :ref:
****************************
The file below demonstrates linking to headings using ``:ref:`` .

This Example File
*****************
{xrst_literal}

{xrst_end heading_example}
"""



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin indent user}

Indentation
###########
If there are a number of spaces (or tabs) before
all of the xrst documentation for a page,
those characters are not included in the xrst output.
This enables one to indent the
xrst so it is grouped with the proper code block in the source.
An error message will result if
you use mix tabs and spaces for the indentation.

Example
*******
:ref:`indent_example-name`, and
:ref:`comment_ch_example@Indent`.

{xrst_end indent}
"""
import re
import xrst
# {xrst_begin remove_indent dev}
# {xrst_comment_ch #}
#
# Remove indentation for a page
# #############################
#
# Arguments
# *********
#
# data_in
# =======
# is the data for this page.
#
# file_name
# =========
# is the input that this page appears in (used for error reporting).
#
# page_name
# =========
# is the name of this page (used for error reporting).
#
# Returns
# *******
#
# data_out
# ========
# is a copy of data_in with the indentation for this section removed.
#
# indent
# ======
# is the white space that was removed from each line (except for empty lines)
#
# {xrst_code py}
def remove_indent(data_in, file_name, page_name) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end remove_indent}
   #
   # len_data
   len_data   = len(data_in)
   #
   # newline_list
   newline_list = xrst.newline_indices(data_in)
   #
   # num_remove
   num_remove = len(data_in)
   for newline in newline_list :
      next_ = newline + 1
      if next_ < len_data and 0 < num_remove :
         ch = data_in[next_]
         while ch in ' \t' and next_ + 1 < len_data :
            next_ += 1
            ch     = data_in[next_]
         if ch not in ' \t\n' :
            num_remove = min(num_remove, next_ - newline - 1)
   #
   # check if there is no indent to remove
   if num_remove == 0 :
      return data_in, ''
   #
   # indent_ch
   line      = 0
   indent_ch = data_in[ newline_list[line] + 1 ]
   while indent_ch == '\n' :
      line += 1
      indent_ch = data_in[ newline_list[line] + 1 ]
   #
   # check for mixing spaces and tabs
   check_ch  = indent_ch + '\n'
   for newline in newline_list :
      next_ = newline + 1
      end   = min( len_data, next_ + num_remove )
      while next_ < end :
         if data_in[next_] not in check_ch :
            msg  = 'mixing both spaces and tabs for '
            msg += 'white space that indents this page.'
            xrst.system_exit(
               msg, file_name=file_name, page_name=page_name
            )
         next_ += 1
   #
   # data_out
   pattern  = re.compile( r'\n' + num_remove * indent_ch )
   data_out = pattern.sub('\n', data_in)
   #
   # indent
   indent = num_remove * indent_ch
   # BEGIN_return
   assert type(data_out) == str
   assert type(indent) == str
   #
   return data_out, indent
   # END_return



# vim:nofixeol
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin testExample}
{xrst_spell
   underbar
}
Test Special Conditions
#######################

Camel Case
**********
This page name uses a capital letter, instead of an underbar,
to separate the words test and example.

Code Command in List
********************
#. The ``\{xrst_code py}`` cannot be on the first line of a list item.
   {xrst_code py}"""
   def factorial(n) :
      if n == 1 :
         return 1
      return n * factorial(n-1)
   """{xrst_code}
#. This is the second item for this list

No Newline at End of File
*************************
This example file does not have a newline at the end.

This Example File
*****************
{xrst_literal}

{xrst_end testExample}
"""


# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
# pattern
pattern = re.compile( r'^\n[ \t]*' )
#
# {xrst_begin add_line_numbers dev}
# {xrst_comment_ch #}
#
# Add Line Numbers to File Data
# #############################
# Add line numbers to a string in a way that is useful for reporting errors
# (for modified versions of string) using line number in the original string.
#
# Arguments
# *********
#
# data_in
# =======
# The original string.  An empty line is a line with just spaces or tabs.
# line_number is the number of newlines before a line plus one; i.e.,
# the first line is number one.
#
# Returns
# *******
#
# data_out
# ========
# The return data_out is a modified version of data_in. The text
#
#  | ``\{xrst_line`` *line_number@*
#
# is added at the end of each non-empty line.
# Spaces and tabs in empty lines are removed (so they are truly empty).
#
# {xrst_code py}
def add_line_numbers(data_in) :
   assert type(data_in) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end add_line_numbers}
   #
   # data_extend
   data_extend = data_in
   if data_extend[-1] != '\n' :
      data_extend += '\n'
   #
   # newline_list, line_start
   newline_list = xrst.newline_indices(data_extend)
   if newline_list[0] == 0 :
      line_start = 2
      newline_list .pop(0)
   else :
      line_start = 1
   #
   # data_out, previous
   data_out     = ""
   previous     = 0
   #
   for i in range( len(newline_list) ) :
      #
      # current
      current = newline_list[i]
      assert previous < current
      #
      # line
      line = data_extend[previous : current]
      #
      # empty_line
      if previous == 0 :
         m_obj = pattern.search( '\n' + line )
         empty_line = m_obj.end() == len(line) + 1
      else :
         m_obj = pattern.search( line )
         empty_line = m_obj.end() == len(line)
      #
      # line
      if empty_line :
         line = '\n'
      else :
         line += '{xrst_line ' + str(i + line_start) + '@'
      #
      # data_out, previous
      data_out  += line
      previous = current
   #
   # data_out
   assert previous == len(data_extend) - 1
   data_out += '\n'
   #
   # BEGIN_return
   assert type(data_out) == str
   return data_out
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
import xrst
# {xrst_begin start_end_file dev}
# {xrst_spell
#     cmd
# }
# {xrst_comment_ch #}
#
# Convert literal command start, end from text to line numbers
# ############################################################
#
# Arguments
# *********
#
# file_cmd
# ========
# is the name of the file where the xrst_literal command appears.
#
# page_name
# =========
# is the name of the page where the xrst_literal command appears.
#
# display_file
# ============
# is the name of the file that we are displaying. If it is not the same as
# file_cmd, then it must have appeared in the xrst_literal command.
#
# cmd_line
# ========
# If file_cmd is equal to display_file, the lines of the file
# between line numbers cmd_line[0] and cmd_line[1] inclusive
# are in the xrst_literal command and are excluded from the search.
#
# start_after
# ===========
# is the starting text. There must be one and only one copy of this text in the
# file (not counting the excluded text). This text has no newlines and cannot
# be empty.  If not, an the error is reported and the program stops.
#
# end_before
# ==========
# is the stopping text. There must be one and only one copy of this text in the
# file (not counting the excluded text). This text has no newlines and cannot
# be empty.  Furthermore, the stopping text must come after the end of the
# starting text. If not, an the error is reported and the program stops.
#
# Returns
# *******
#
# start_line
# ==========
# is the line number where start_after appears.
#
# end_line
# ========
# is the line number where end_before appears.
#
# {xrst_code py}
def start_end_file(
   file_cmd,
   page_name,
   display_file,
   cmd_line,
   start_after,
   end_before
) :
   assert type(file_cmd) == str
   assert type(page_name) == str
   assert type(display_file) == str
   assert type(cmd_line[0]) == int
   assert type(cmd_line[1]) == int
   assert cmd_line[0] <= cmd_line[1]
   assert type(start_after) == str
   assert type(end_before) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end start_end_file}
   # ------------------------------------------------------------------------
   # exclude_line
   if file_cmd == display_file :
      exclude_line = cmd_line
   else :
      exclude_line = (0, 0)
   #
   # msg
   msg  = f'in literal command:'
   #
   if start_after == '' :
      msg += ' start_after is empty'
      xrst.system_exit(msg,
         file_name=file_cmd, page_name=page_name, line = cmd_line[0]
      )
   if end_before == '' :
      msg += ' end_before is empty'
      xrst.system_exit(msg,
         file_name=file_cmd, page_name=page_name, line = cmd_line[0]
      )
   if 0 <= start_after.find('\n') :
      msg += ' a newline appears in start_after'
      xrst.system_exit(msg,
         file_name=file_cmd, page_name=page_name, line = cmd_line[0]
      )
   if 0 <= end_before.find('\n') :
      msg += ' a newline appears in end_before'
      xrst.system_exit(msg,
         file_name=file_cmd, page_name=page_name, line = cmd_line[0]
      )
   #
   # data
   file_obj  = open(display_file, 'r')
   data      = file_obj.read()
   file_obj.close()
   #
   # start_line
   start_index = data.find(start_after)
   count = 0
   while 0 <= start_index :
      line = data[: start_index].count('\n') + 1
      if  line < exclude_line[0] or exclude_line[1] < line :
         start_line = line
         count      = count + 1
      start_index = data.find(start_after, start_index + len(start_after) )
   if count != 1 :
      msg += f'\nstart_after   = "{start_after}"'
      msg += f'\ndisplay_file  =  {display_file}'
      msg += f'\nfound {count} matches expected 1'
      if file_cmd == display_file :
         msg += ' not counting the literal command'
      xrst.system_exit(msg,
         file_name=file_cmd, page_name=page_name, line = cmd_line[0]
      )
   #
   # end_line
   stop_index = data.find(end_before)
   count = 0
   while 0 <= stop_index :
      line = data[: stop_index].count('\n') + 1
      if  line < exclude_line[0] or exclude_line[1] < line :
         end_line = line
         count     = count + 1
      stop_index = data.find(end_before, stop_index + len(end_before) )
   if count != 1 :
      msg += f'\nend_before   = "{end_before}"'
      msg += f'\ndisplay_file =  {display_file}'
      msg += f'\nfound {count} matches expected 1'
      if file_cmd == display_file :
         msg += ' not counting the literal command'
      xrst.system_exit(msg,
         file_name=file_cmd, page_name=page_name, line = cmd_line[0]
      )
   # ------------------------------------------------------------------------
   # BEGIN_return
   assert type(start_line) == int
   assert type(end_line) == int
   #
   return start_line, end_line
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
import os
import re
import subprocess
# ----------------------------------------------------------------------------
def get_index_page_name() :
   file_obj  = open('.readthedocs.yaml', 'r')
   file_data = file_obj.read()
   pattern   = r'\n *--index_page_name *([-._a-zA-Z0-9]*)'
   m_obj     = re.search(pattern, file_data)
   file_obj.close()
   return m_obj.group(1)
# ----------------------------------------------------------------------------
def get_rst_directory() :
   file_obj  = open('xrst.toml', 'r')
   file_data = file_obj.read()
   pattern   = r"\nrst_directory *= *'([^']*)'"
   m_obj     = re.search(pattern, file_data)
   file_obj.close()
   return m_obj.group(1)
# ----------------------------------------------------------------------------
def run_xrst() :
   index_page_name = get_index_page_name()
   #
   # This command should be the same as the one in bin/check_xrst.sh
   # that's used to keep the test_rst directory up to date.
   command = [
      'python3', '-m', 'xrst',
      '--local_toc',
      '--rst_only',
      '--config_file',     'xrst.toml',
      '--index_page_name', index_page_name,
      '--group_list',      'default', 'user', 'dev',
      '--html_theme',      'sphinx_rtd_theme',
   ]
   result = subprocess.run(command)
   assert result.returncode == 0
# ----------------------------------------------------------------------------
def run_test() :
   run_xrst()
   #
   # rst_list
   rst_directory = get_rst_directory()
   rst_list      = list()
   for entry in os.listdir(rst_directory) :
      if entry.endswith('.rst') :
         rst_list.append(entry)
   rst_list = sorted(rst_list)
   #
   # check_list
   check_list = sorted( os.listdir('test_rst') )
   #
   # rst_index, check_index
   rst_index   = 0
   check_index = 0
   while rst_index < len(rst_list) and check_index < len(check_list) :
      #
      # rst_name, check_name
      rst_name   = rst_list[rst_index]
      check_name = check_list[check_index]
      if rst_name < check_name :
         rst_index += 1
         assert False, f'{rst_name} is in {rst_directory} but not test_rst'
      elif check_name < rst_name :
         check_index += 1
         assert False, f'{check_name} is in test_rst but not {rst_directory}'
      else :
         #
         # rst_index, check_index
         rst_index     += 1
         check_index   += 1
         #
         # rst_file_data, check_file_data
         rst_file      = f'{rst_directory}/{rst_name}'
         check_file    = f'test_rst/{check_name}'
         #
         rst_file_obj   = open(rst_file, 'r')
         check_file_obj = open(check_file, 'r')
         #
         rst_data      = rst_file_obj.read()
         check_data    = check_file_obj.read()
         #
         rst_file_obj.close()
         check_file_obj.close()
         #
         if rst_data == check_data :
            print( f'{rst_name}: OK' )
         else :
            msg = f'{rst_file} is different from {check_file}'
            breakpoint()
            assert False, msg
# ----------------------------------------------------------------------------
def test_rst() :
   if not os.path.exists('xrst.toml') :
      assert False, 'test_rst.py: must start in the xrst top source directory'
   else :
      run_test()
# ----------------------------------------------------------------------------
if __name__ == '__main__' :
   test_rst()



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
def factorial(n) :
   """
   {xrst_begin indent_example}
   {xrst_spell
      docstring
   }

   Indent Example
   ##############

   Discussion
   **********
   The file below demonstrates indenting an entire xrst page.
   Note that underling headings works even though it is indented.

   Python Docstring
   ****************
   This example input is a python docstring for the factorial function
   defined in this file, but it is documenting indentation instead
   of the function. See :ref:`docstring_example-name` for an alternative
   way to construct a docstring.

   This Example File
   *****************
   {xrst_literal}

   {xrst_end indent_example}
   """
   if n == 1 :
      return 1
   return n * factorial(n-1)



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
pattern_start_literal = re.compile( r'\n::{xrst_line [0-9]+@\n\n' )
pattern_end_literal   = re.compile( r'\n[^ \t]')
pattern_declare = re.compile(
   r'\n\.\. _([^:\n]+):([^\n]*)\{xrst_line ([0-9]+)@'
)
pattern_use = re.compile(
   r'`([^<\n]+)(<[^>\n]*>)`_[^\n]*\{xrst_line ([0-9]+)@'
)
# ----------------------------------------------------------------------------
# {xrst_begin sphinx_label dev}
#
# Get Labels Declared Using Sphinx Commands
# #########################################
#
# {xrst_comment_ch #}
#
# data_in
# *******
# is the data for one page.
# Line numbers have been added to this data; see
# :ref:`add_line_numbers-name` .
#
# file_name
# *********
# is the name of the xrst input file corresponding to data_in
# (only used for error reporting).
#
# page_name
# *********
# is the page name corresponding to data_in
# (only used for error reporting).
#
# external_line
# *************
# For each label declared in data_in using sphinx commands,
# and that links to an external web site,
# *external_line* [ *label*.lower() ] is the line number in
# *file_name* where the label was defined.
#
# internal_line
# *************
# For each label declared in data_in using sphinx commands,
# and that links to a page in this web site,
# *internal* [ *label* ] is the line number in
# *file_name* where the label was defined.
#
# Errors
# ******
# If two or external labels have the same lower case value,
# an error is reported using :ref:`system_exit-name` .
#
# {xrst_code py}
# external_line, internal_line =
def sphinx_label(data_in, file_name, page_name) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   # {xrst_code}
   # {xrst_literal
   #     BEGIN_return
   #     END_return
   # }
   # {xrst_end sphinx_label}
   #
   # external_line, internal_line
   external_line  = dict()
   internal_line  = dict()
   #
   # data
   data = data_in
   #
   # m_start
   # remove literal blocks because no labels defined there
   m_start = pattern_start_literal.search(data)
   while m_start :
      #
      # m_end
      m_end = pattern_end_literal.search(data_in, m_start.end() )
      #
      # data, m_start
      if m_end == None :
         data    = data[: m_start.start() ]
         m_start = None
      else :
         data_before = data[: m_start.start() ]
         data_after  = data[m_end.end() : ]
         data        = data_before + data_after
         m_start     = pattern_start_literal.search(data, len(data_before) )
   #
   # pattern
   for pattern in [ pattern_declare, pattern_use ] :
      #
      # m_label
      m_label    = pattern.search(data)
      while m_label :
         #
         # label_lower, line
         label       = m_label.group(1).strip(' ')
         label_lower = label.lower()
         destination = m_label.group(2).strip(' ')
         line        = m_label.group(3)
         #
         for ending in [ '-name', '-title' ] :
            if label_lower.endswith(ending) :
               msg  = f'Labels that end with {ending} are not allowed\n'
               msg += 'because they are automatically generated by xrst\n'
               msg += f'label.lower() = {label_lower}'
               xrst.system_exit(msg,
                  file_name = file_name, page_name = page_name, line = line
               )
         if '@' in label_lower :
            msg  = f'Labels that use the @ character are not allowed\n'
            msg += 'because they are automatically generated by xrst\n'
            msg += f'label.lower() = {label_lower}'
            xrst.system_exit(msg,
               file_name = file_name, page_name = page_name, line = line
            )
         if label_lower == page_name :
            msg  = 'A label has the same lower case value as the page_name\n'
            msg += f'label.lower() = {label_lower}'
            xrst.system_exit(
               msg, file_name = file_name, page_name = page_name, line = line
            )
         #
         if label_lower in external_line :
            previous_line = external_line[label_lower]
            msg  = 'A previous label has the same lower case value\n'
            msg += f'label.lower() = {label_lower}\n'
            msg += f'previous_line = {previous_line}'
            xrst.system_exit(
               msg, file_name = file_name, page_name = page_name, line = line
            )
         #
         # external_line, internal_line
         external =  destination != ''
         if external :
            external_line[label_lower] = line
         else :
            internal_line[label] = line
         #
         # m_label
         m_label = pattern.search(data, m_label.end())
   # BEGIN_return
   for result in [ external_line, internal_line ] :
      assert type(result) == dict
      for key in result.keys() :
         assert type( result[key] ) == str
   #
   return external_line, internal_line
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
# This is a temporary fix and duplicates information in pyproject.toml.
# It will be removed when seutptools gets the pyproject.toml dependencies right
from setuptools import setup

setup(
   name             = 'xrst',
   version          = '2023.1.10',
   description      = 'Extract RST files from source code and run Sphinx',
   license          = 'GPL-3.0-or-later',
   keywords         = 'sphinx rst documentation source',
   url              = 'https://github.com/bradbell/xrst',
   packages         = ['xrst'],
   long_description = 'readme.md',
   classifiers      = [
      'Topic :: Documentation :: Sphinx',
      'Development Status :: 5 - Production/Stable',
      'License :: OSI Approved',
      'Operating System :: OS Independent',
      'Environment :: Console',
      'Intended Audience :: Developers',
      'Natural Language :: English',
   ],
   install_requires = [
      'sphinx', 'toml',                                # required
      'build', 'pytest',                               # ned for testing xrst
      'pyenchant', 'pyspellchecker',                   # need the one you use
      'furo', 'sphinx-rtd-theme', 'sphinx-book-theme', # need the ones you use
   ],
   entry_points = {
      'console_scripts' : [ 'xrst = xrst:run_xrst' ]
   }
)



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin suspend_cmd user}

Suspend and Resume Commands
###########################

Syntax
******
- ``\{xrst_suspend}``
- ``\{xrst_resume}``

Purpose
*******
It is possible to suspend (resume) the xrst extraction during a page.
One begins (ends) the suspension with a line that only contains spaces,
tabs and a suspend command (resume command).
Note that this will also suspend all other xrst processing; e.g.,
spell checking.

Example
*******
:ref:`suspend_example-name`

{xrst_end suspend_cmd}
"""
# ----------------------------------------------------------------------------
import re
import xrst
#
# pattern_suspend, pattern_resume
pattern_suspend = re.compile(
   r'[^\\]{xrst_suspend}'
)
pattern_resume  = re.compile(
   r'[^\\]{xrst_resume}'
)
# {xrst_begin suspend_cmd_dev dev}
# {xrst_comment_ch #}
#
# Remove text specified by suspend / resume pairs
# ###############################################
#
# Arguments
# *********
#
# data_in
# =======
# is the data for this page.
#
# file_name
# =========
# is the input file corresponding to this page.
#
# page_name
# =========
# is the name of this page.
#
# Returns
# *******
#
# data_out
# ========
# The return data_out is a copy of data_in except that the text between
# and including each suspend / resume pair has been removed.
#
# {xrst_code py}
def suspend_command(data_in, file_name, page_name) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end suspend_cmd_dev}
   #
   # data_out
   data_out = data_in
   #
   # m_suspend
   m_suspend  = pattern_suspend.search(data_out)
   while m_suspend != None :
      #
      # suspend_stat, suspend_end
      suspend_start = m_suspend.start() + 1
      suspend_end   = m_suspend.end()
      #
      # m_resume
      m_resume      = pattern_resume.search(data_out, suspend_end)
      if m_resume == None :
         msg  = 'There is a suspend command without a '
         msg += 'corresponding resume commannd.'
         xrst.system_exit(msg,
            file_name=file_name,
            page_name=page_name,
            m_obj=m_suspend,
            data=data_out
         )
      # resume_start, resume_end
      resume_start = m_resume.start() + 1
      resume_end   = m_resume.end()
      #
      # m_obj
      m_obj = pattern_suspend.search(data_out, suspend_end)
      if m_obj != None :
         if m_obj.start() < resume_end :
            msg  = 'There are two suspend commands without a '
            msg += 'resume command between them.'
            xrst.system_exit(msg,
               file_name=file_name,
               page_name=page_name,
               m_obj=m_obj,
               data=data_rest
            )
      #
      # data_out
      data_tmp  = data_out[: suspend_start]
      data_tmp += data_out[resume_end : ]
      data_out  = data_tmp
      #
      # m_suspend
      m_suspend = pattern_suspend.search(data_out)
   #
   # check_syntax_error
   for command_name in [ 'suspend', 'resume' ] :
      xrst.check_syntax_error(
         command_name  = command_name,
         data          = data_out,
         file_name     = file_name,
         page_name     = page_name,
      )
   # BEGIN_return
   assert type(data_out) == str
   return data_out
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin get_spell_checker dev}
{xrst_spell
   pyspellchecker
}

Get A Spell Checker Object
##########################

Syntax
******

| *spell_checker* = xrst.get_spell_checker(*local_words*, *package*)
| *known*   = *spell_checker*.known ( *word* )
| *suggest* = *spell_checker*.suggest ( *word* )

local_words
***********
is a list of words
(each word is a non-empty str)
that get added to the dictionary for this spell checker.
No need to add single letter words because they are considered correct
by spell_command routine.
}

package
*******
is an str equal to 'pyspellchecker' or 'enchant' .

word
****
is a word the we are either checking to see if it is correct,
or looking for a suggested spelling for.

spell_checker
*************
is the spell checking object.

known
*****
is True (False) if *word* is (is not) a correctly spelled word.

suggest
*******
if *word* is correctly spelled, *suggest* is equal to *word* .
Otherwise if *suggest* is not None, it is a suggestion
for correcting the spelling.
If *suggest* is None, the spell checker does not have a suggestion
for correcting the spelling of *word*. .


{xrst_end get_spell_checker}
'''
# -----------------------------------------------------------------------------
# remove_from_dictionary
# list of words that, if they are in the dictionary, are removed
remove_from_dictionary = [
   # BEGIN_SORT_THIS_LINE_PLUS_1
   'af',
   'anl',
   'ap',
   'av',
   'bnd',
   'bv',
   'cg',
   'conf',
   'cpp',
   'dep',
   'dir',
   'dv',
   'exp',
   'gcc',
   'hes',
   'hess',
   'ind',
   'jac',
   'len',
   'mcs',
   'meas',
   'nc',
   'nd',
   'nr',
   'op',
   'prt',
   'ptr',
   'rc',
   'rel',
   'sim',
   'std',
   'tbl',
   'thier',
   'var',
   'vec',
   'xp',
   'yi',
   # END_SORT_THIS_LINE_MINUS_1
]
# -----------------------------------------------------------------------------
# add_to_dictionary
# list of
add_to_dictionary = [
   # BEGIN_SORT_THIS_LINE_PLUS_1
   'aborts',
   'asymptotic',
   'configurable',
   'covariate',
   'covariates',
   'debug',
   'deprecated',
   'destructor',
   'exponentiation',
   'hessians',
   'html',
   'identifiability',
   'indenting',
   'initialization',
   'initialize',
   'initialized',
   'integrand',
   'integrands',
   'invertible',
   'jacobian',
   'jacobians',
   'likelihoods',
   'messaging',
   'modeled',
   'modeling',
   'multipliers',
   'optimizes',
   'partials',
   'tex',
   'piecewise',
   'subdirectory',
   'unary',
   'unicode',
   'verbose',
   'wiki',
   'wikipedia',
   'xrst',
   # END_SORT_THIS_LINE_MINUS_1
]
# -----------------------------------------------------------------------------
class py_spell_checker :
   #
   # self
   def __init__(self, local_words) :
      assert type(local_words) == list
      for word in local_words :
         assert type(word) == str
      #
      # spellchecker
      import spellchecker
      #
      # checker
      checker = spellchecker.SpellChecker(distance=1)
      #
      # checker
      remove_list = checker.known( remove_from_dictionary )
      checker.word_frequency.remove_words(remove_list)
      #
      # checker
      checker.word_frequency.load_words(add_to_dictionary)
      checker.word_frequency.load_words(local_words)
      #
      self.checker = checker
   #
   # ok
   def known(self, word) :
      assert type(word) == str
      #
      return len( self.checker.unknown( [word] ) ) == 0
   #
   # suggest
   def suggest(self, word) :
      assert type(word) == str
      #
      return self.checker.correction(word)
# -----------------------------------------------------------------------------
class enchant_spell_checker :
   #
   # self
   def __init__(self, local_words) :
      assert type(local_words) == list
      for word in local_words :
         assert type(word) == str
      #
      # enchant
      import enchant
      #
      # checker
      checker =  enchant.Dict("en_US")
      #
      # checker
      for word in remove_from_dictionary :
         checker.remove(word)
      #
      # checker
      for word in add_to_dictionary + local_words :
         checker.add(word)
      #
      self.checker = checker
   #
   # ok
   def known(self, word) :
      assert type(word) == str
      #
      return self.checker.check(word)
   #
   # suggest
   def suggest(self, word) :
      assert type(word) == str
      #
      suggest_list =  self.checker.suggest(word)
      if len(suggest_list) == 0 :
         return None
      return suggest_list[0]
# -----------------------------------------------------------------------------
def get_spell_checker(local_words, package) :
   if package == 'pyspellchecker' :
      return py_spell_checker(local_words)
   else :
      assert package == 'pyenchant'
      return enchant_spell_checker(local_words)



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
#
pattern_error = re.compile( r'{xrst_line [0-9]+@[^\n]' )
# {xrst_begin remove_line_numbers dev}
# {xrst_comment_ch #}
#
# Remove the number numbers
# #########################
#
# Arguments
# *********
#
# data_in
# =======
# is a string with line number markers added by :ref:`add_line_numbers-name` .
# These lines number markers have the form:
#
#     ``{xrst_line`` *line_number* ``@`` .
#
# Returns
# *******
#
# data_out
# ========
# The return data_out is a copy of data_in with the
# line number markers removed.
#
# line_pair
# =========
# The second return line_pair is a list of two element tuples.
#
# -   The first element is the line number in data_out corresponding to
#     the line number marker that was removed.
#     These line numbers, in data_out, do not count
#     lines that only contain ``{xrst_before_title}`` .
#
# -   The second element is the *line_number*, in the line number marker,
#     that was removed.
#
# -   The data_out line numbers are in increasing order and
#     the maker line numbers are non-decreasing.
#
# {xrst_code py}
def remove_line_numbers(data_in) :
   assert type(data_in) == str
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end remove_line_numbers}
   #
   # m_error
   m_error = pattern_error.search(data_in)
   if m_error :
      start = max(m_error.start() - 50, 0)
      end   = min(m_error.end() + 50, len(data_in))
      msg   = 'Program error: Line number tracking is confused:\n'
      msg  += '\nText before the bad line number ='
      msg  += '\n---------------------------------\n'
      msg  +=  data_in[start : m_error.start()]
      msg  += '\nText after the bad line number ='
      msg  += '\n--------------------------------\n'
      msg  +=  data_in[m_error.end() :  end]
      xrst.system_exit(msg)
   #
   # previous_end
   # index of the end of the previous match
   previous_end  = 0
   #
   # line_out
   # index of next line in data_out
   line_out  = 1
   #
   # data_out, line_pair
   data_out  = ''
   line_pair = list()
   #
   # data_out, line_pair
   for m_obj in xrst.pattern['line'].finditer(data_in) :
      #
      # start of this match
      this_start = m_obj.start()
      #
      # before
      # character from end of previous match to start of this match
      before = data_in[previous_end  : this_start]
      #
      line_match = m_obj.group(1)
      line_out  += before.count('\n')
      line_out  -= before.count('{xrst_before_title}\n')

      line_pair.append( ( line_out, int(line_match) ) )
      data_out += before
      #
      previous_end = m_obj.end()
   #
   # data_out
   data_out += data_in[previous_end  :]
   #
   # BEGIN_return
   assert type(data_out) == str
   assert type(line_pair) == list
   if 0 < len(line_pair) :
      assert type(line_pair[0]) == tuple
      assert type(line_pair[0][0]) == int
      assert type(line_pair[0][1]) == int
   #
   return data_out, line_pair
   # END_return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
"""
{xrst_begin heading_links user}

Heading Cross Reference Links
#############################

Index
*****
For each word in a heading,
a link is included in the index from the word to the heading.
In addition, each word is added to the html keyword meta data
next to the page heading.

Labels
******
A cross reference label is defined for linking
from anywhere to a heading. The details of how to use
these labels are described below.

Level Zero
==========
Each :ref:`page<begin_cmd@page>` can have only one header at
the first level which is a title for the page.

page_name
---------
The input below will display the page name as the linking text:

  ``:ref:`` \` *page_name* ``-name`` \`

page_title
----------
The input below will display the page title as the linking text:

    ``:ref:`` \` *page_name* ``-title`` \`

Linking Text
------------
You can also explicitly choose the linking text using:

   ``:ref:`` \` *linking_text* ``<`` *page_name* ``-name>`` \`


Other Levels
============
The label for linking a heading that is not at level zero is the label
for the heading directly above it plus an at sign character :code:`@`,
plus the conversion for this heading.
These labels use the *page_name* for level zero,
without the ``-name`` or ``--title`` at the end.

Heading@To@Label
================
The conversion of a heading to a label
removes all backslashes ``\`` and changes at signs ``@``
to dashes ``-``.

For example, the label for the heading above is

   :ref:`heading_links@Labels@Heading-To-Label
   <heading_links@Labels@Heading-To-Label>`

The label corresponding to a header is used to reference the heading
using the ``:ref:`` role.

Label To Anchor
===============
There is a further conversion to create the
HTML anchor corresponding to a label.  To be specific:

1. The anchor is converted to lower case.
3. Characters that are not letters or decimal digits are converted to dashes.
4. Multiple dashes are converted to one dash.
5. The beginning of the anchor is trimmed until a letter is reached.
6. The end of the anchor is trimmed until a letter or digit is reached.

If for one page, these anchors are not unique, xrst reports an error.


Discussion
==========
#. Note that for level zero one uses the *page_name* and not the
   title; e.g., in the example above one uses ``heading_links``
   and not ``Heading Cross Reference Links`` .
#. The ``@`` and not ``.`` character is used to separate levels
   because the ``.`` character is often used in titles and
   page names; e.g. :ref:`auto_file@conf.py`.
#. The xrst automatically generated labels end in ``-name`` , ``-title`` ,
   or have a ``@`` character in them. Other labels, that you create using
   rst commands, should not satisfy this condition
   (and hence are easy to distinguish).
#. Including all the levels above a heading in its label may seem verbose,
   but it avoids ambiguity when the same heading appears twice in one page.
   For example, this link to the project name
   :ref:`config_file@project_name@Default`
   which is one of many Default headings on that page.
   This also helps keep the links up to date.
   If a heading changes, all the links to that heading, and all the
   headings below it,  will break.  This identifies the links that should be
   checked to make sure they are still valid.
#. It is an error for two headings have the same HTML anchor.
   This makes the html links to a heading valid as long as its label
   does not change. This is useful when posting the answer to a questions
   using a link to a particular heading.

Example
*******
:ref:`heading_example-name`

{xrst_end heading_links}
-------------------------------------------------------------------------------
"""
import re
import xrst
# -----------------------------------------------------------------------------
def check_anchor( label, line, file_name, page_name, previous_anchor) :
   assert type(label) == str
   assert type(line) == str
   assert type(file_name) == str
   assert type(previous_anchor) == dict
   #
   # anchor
   # convert to the following pattern: [a-z](-?[a-z0-9]+)*
   anchor = label.lower()
   anchor = re.sub( r'[^a-z0-9]', '-', anchor)
   anchor = re.sub( r'-+',        '-', anchor)
   anchor = re.sub( r'-$',        '',  anchor)
   anchor = re.sub( r'^[^a-z]*',  '',  anchor)
   if anchor == '' :
      msg  = 'The anchor correspnding to a header is empty.\n'
      msg += f'label = {label}'
      xrst.system_exit(
         msg, file_name = file_name, page_name = page_name, line = line,
      )
   #
   # check for duplicate anchor
   if anchor in previous_anchor :
      previous_line  = previous_anchor[anchor]['line']
      previous_label = previous_anchor[anchor]['label']
      msg  = 'A previous header has the same HTML anchor.\n'
      msg += f'label          = {label}\n'
      msg += f'previous_label = {previous_label}\n'
      msg += f'anchor         = {anchor}\n'
      msg += f'previous_line  = {previous_line}'
      xrst.system_exit(
         msg, file_name = file_name, page_name = page_name, line = line,
      )
      assert False, msg
   #
   # previous_anchor
   previous_anchor[anchor] = { 'line' : line, 'label' : label }
# -----------------------------------------------------------------------------
# {xrst_begin process_headings dev}
# {xrst_spell
#     bool
#     fullmatch
#     overline
#     toc
# }
# {xrst_comment_ch #}
#
# Add labels and index entries for headings
# #########################################
#
# Arguments
# *********
#
# local_toc
# =========
# is the xrst command line local_toc setting.
#
# data_in
# =======
# contains the data for a page before the headings are processed.
#
# file_name
# =========
# name of the file that contains the input data for this page.
# This is only used for error reporting.
#
# page_name
# =========
# is the name of this page.
#
# not_in_index_list
# =================
# is a list of compiled regular expressions. If pattern is in this list,
# *word* is a lower case version of a word in the heading text, and
# pattern.fullmatch( *word* ) returns a match, an index entry is not
# generated for word.
#
# Returns
# *******
#
# data_out
# ========
# is a copy of data_in with the following extra command added:
#
#  #. The index entries, and meta keyword entries (same as index),
#     and the :ref:`heading_links@Labels` for this page.
#  #. The command \\n{xrst_before_title} is placed directly before the
#     first heading for this page; i.e. its title.
#     This is makes it easy to add the page number to the heading text.
#
# page_title
# ==========
# This is the heading text in the first heading for this page.
# There can only be one heading at this level.
#
# pseudo_heading
# ==============
# This is an automatically generated heading for this page. It is intended
# to come before the page_title heading.
# It has three lines each terminated by a newline:
#
#  1. an overline line
#  2. heading text line for this page title
#  3. an underline line
#
# {xrst_code py}
def process_headings(
      local_toc, data_in, file_name, page_name, not_in_index_list
) :
   assert type(local_toc) == bool
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   assert type(not_in_index_list) == list
   # {xrst_code}
   # {xrst_literal
   #  BEGIN_return
   #  END_return
   # }
   # {xrst_end process_headings}
   #
   # pattern_colon_in_label
   pattern_colon_space = re.compile( r':(\s)' )
   #
   # previous_anchor
   previous_anchor = dict()
   #
   # external_line, internal_line
   external_line, internal_line = xrst.sphinx_label(
      data_in, file_name, page_name
   )
   #
   # previous_anchor
   for label in internal_line :
      line = internal_line[label]
      check_anchor(
         label, line, file_name, page_name, previous_anchor, exteranl_line
      )
   #
   # data_out
   data_out = data_in
   #
   # punctuation
   punctuation      = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
   assert len(punctuation) == 34 - 2 # two escape sequences
   #
   # overline_used
   overline_used = set()
   #
   # found_level_one_heading
   found_level_one_heading = False
   #
   # heading_list, heading_index, heading_text, underline_text
   heading_list     = list()
   data_index       = 0
   heading_index, heading_text, underline_text = \
      xrst.next_heading(data_out, data_index, file_name, page_name)
   #
   while 0 <= heading_index :
      if 0 < heading_index :
         assert data_out[heading_index-1] == '\n'
      #
      # m_line
      m_line = xrst.pattern['line'].search(data_out, heading_index)
      #
      # overline
      index = m_line.start()
      overline = underline_text == data_out[heading_index : index]
      #
      # character
      character = underline_text[0]
      #
      # heading
      heading   = {
         'overline' : overline,
         'character': character,
         'text':      heading_text
      }
      #
      # underline_end
      underline_end = data_out.find('\n', heading_index)
      underline_end = data_out.find('\n', underline_end+1)
      if overline :
         underline_end = data_out.find('\n', underline_end+1)
      assert data_out[underline_end] == '\n'
      #
      # overline_used
      if overline :
         overline_used.add(character)
      #
      # heading_list
      if len( heading_list ) == 0 :
         # first heading in this page
         heading_list.append( heading )
      else :
         # level_zero
         level_zero = overline == heading_list[0]['overline']
         if level_zero :
            level_zero = character == heading_list[0]['character']
         if level_zero :
            msg = 'There are multiple titles for this page'
            xrst.system_exit(
               msg,
               file_name = file_name,
               page_name = page_name,
               line      = m_line.group(1),
            )
         #
         # found_level
         found_level = False
         level       = 1
         while level < len(heading_list) and not found_level :
            found_level = overline == heading_list[level]['overline']
            if found_level :
               found_level = character == heading_list[level]['character']
            if found_level :
               #
               # heading_list
               heading_list = heading_list[: level ]
               heading_list.append(heading)
            else :
               level += 1
         #
         # heading_list
         if not found_level :
            # this heading at a higher level
            heading_list.append( heading )
      #
      # label
      label = None
      for level in range( len(heading_list) ) :
         if level == 0 :
            assert page_name == page_name.replace('\\', '').replace('@', '_')
            #
            # label
            if len(heading_list) == 1 :
               label = page_name + '-title'
            else :
               label = page_name
         else :
            conversion  = heading_list[level]['text']
            conversion  = conversion.replace('\\', '')
            conversion  = conversion.replace('@',  '-')
            conversion  = pattern_colon_space.sub( '\\:\\1', conversion)
            label      += '@' + conversion
      #
      # label
      if label.endswith(':') :
         label = label[:-1] + '\\:'
      if label.startswith('_') :
         label = '\\' + label
      #
      # check external labels
      if label.lower() in external_line :
         other_line = external_line[ label.lower() ]
         msg  = 'This label has same lower case representation as anotheer\n'
         msg += 'other line = {other_line}'
         xrst.system_exit(msg, file_name = file_name, page_name = page_name)
      #
      # check_anchor
      line = m_line.group(1)
      check_anchor(label, line, file_name, page_name, previous_anchor)
      #
      # index_entries
      if len(heading_list) == 1 :
         index_entries = page_name
      else :
         index_entries = ''
      for word in heading_list[-1]['text'].lower().split() :
         skip = False
         for pattern in not_in_index_list :
            m_obj = pattern.fullmatch(word)
            if m_obj :
               skip = True
         if not skip :
            if index_entries == '' :
               index_entries = word
            else :
               index_entries += ', ' + word
      #
      # data_tmp
      # data that comes before this heading
      data_tmp   = data_out[: heading_index]
      #
      # data_tmp
      # If first level one heading and sphinx_rtd_theme,
      # put jump table command before heading
      if len(heading_list) == 2 and not found_level_one_heading :
         found_level_one_heading = True
         if local_toc :
            data_tmp += '\n.. contents::\n'
            data_tmp += 3 * ' ' + ':local:\n\n'
      #
      # data_tmp
      # add sphnix keyword, index, and label commnds
      cmd  = ''
      if index_entries != '' :
            cmd += '.. meta::\n'
            cmd += 3 * ' ' + ':keywords: ' + index_entries + '\n\n'
            cmd += '.. index:: '           + index_entries + '\n\n'
      cmd += '.. _' + label + ':\n\n'
      data_tmp  += cmd
      #
      # data_tmp
      # If level zero, put page number command just before heading
      if len(heading_list) == 1 :
         data_tmp += '{xrst_before_title}\n'
      #
      # data_tmp
      # add data from stat to end of heading
      assert data_out[underline_end] == '\n'
      data_tmp  += data_out[heading_index : underline_end]
      #
      # data_out
      data_right = data_out[underline_end : ]
      data_out   = data_tmp + data_right
      #
      # next heading
      data_index = len(data_tmp) + 1
      heading_index, heading_text, underline_text = \
         xrst.next_heading(data_out, data_index, file_name, page_name)
   #
   if len(heading_list) == 0 :
      msg = 'There are no headings in this page'
      xrst.system_exit(msg, file_name=file_name, page_name=page_name)
   #
   # pseudo_heading
   i = 0
   while punctuation[i] in overline_used :
      i += 1
      if i == len(punctuation) :
         msg  = 'more than ' + len(punctuation) - 1
         msg += ' overlined heading levels'
         xrst.system_exit(
            msg, file_name=file_name, page_name=page_name
         )
   line           = len(page_name) * punctuation[i] + '\n'
   pseudo_heading = line + page_name + '\n' + line + '\n'
   #
   # page_title
   page_title = heading_list[0]['text']
   #
   # BEGIN_return
   assert type(data_out) == str
   assert type(page_title) == str
   assert type(pseudo_heading) == str
   #
   return data_out, page_title, pseudo_heading
   # END_return
